"""Data objects pertaining to accounts."""

from __future__ import annotations

import attr
from terra_proto.cosmos.vesting.v1beta1 import (
    BaseVestingAccount as BaseVestingAccount_pb,
)

from terra_sdk.core import Coins
from terra_sdk.util.json import JSONSerializable

from ...public_key import PublicKey
from .base_account import BaseAccount

__all__ = ["BaseVestingAccount"]


@attr.s
class BaseVestingAccount(JSONSerializable):
    """Stores information about an account with vesting."""

    base_account: BaseAccount = attr.ib()

    original_vesting: Coins = attr.ib(converter=Coins)
    """"""

    delegated_free: Coins = attr.ib(converter=Coins)
    """"""

    delegated_vesting: Coins = attr.ib(converter=Coins)
    """"""

    end_time: int = attr.ib(converter=int)
    """"""

    type_amino = "cosmos-sdk/BaseVestingAccount"
    type_url = "/cosmos.vesting.v1beta1.BaseVestingAccount"

    def get_sequence(self) -> int:
        return self.base_account.get_sequence()

    def get_account_number(self) -> int:
        return self.base_account.get_account_number()

    def get_public_key(self) -> PublicKey:
        return self.base_account.get_public_key()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "base_account": self.base_account.to_amino(),
                "original_vesting": self.original_vesting.to_amino(),
                "delegated_free": self.delegated_free.to_amino(),
                "delegated_vesting": self.delegated_vesting.to_amino(),
                "end_time": str(self.end_time),
            },
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "base_account": self.base_account.to_data(),
            "original_vesting": self.original_vesting.to_data(),
            "delegated_free": self.delegated_free.to_data(),
            "delegated_vesting": self.delegated_vesting.to_data(),
            "end_time": str(self.end_time),
        }

    @classmethod
    def from_amino(cls, amino: dict) -> BaseVestingAccount:
        amino = amino["value"] if "value" in amino else amino
        return cls(
            base_account=BaseAccount.from_amino(amino["base_account"]),
            original_vesting=Coins.from_amino(amino["original_vesting"])
            if amino["original_vesting"]
            else None,
            delegated_free=Coins.from_amino(amino["delegated_free"])
            if amino["delegated_free"]
            else None,
            delegated_vesting=Coins.from_amino(amino["delegated_vesting"])
            if amino["delegated_vesting"]
            else None,
            end_time=amino["end_time"],
        )

    @classmethod
    def from_data(cls, data: dict) -> BaseVestingAccount:
        return cls(
            base_account=BaseAccount.from_data(data["base_account"]),
            original_vesting=Coins.from_data(data["original_vesting"]),
            delegated_free=Coins.from_data(data["delegated_free"]),
            delegated_vesting=Coins.from_data(data["delegated_vesting"]),
            end_time=data["end_time"],
        )

    @classmethod
    def from_proto(cls, proto: BaseVestingAccount_pb) -> BaseVestingAccount:
        return cls(
            base_account=BaseAccount.from_proto(proto.base_account),
            original_vesting=Coins.from_proto(proto.original_vesting),
            delegated_free=Coins.from_proto(proto.delegated_free),
            delegated_vesting=Coins.from_proto(proto.delegated_vesting),
            end_time=proto.end_time,
        )

    def to_proto(self) -> BaseVestingAccount_pb:
        return BaseVestingAccount_pb(
            base_account=self.base_account.to_proto(),
            original_vesting=self.original_vesting.to_proto(),
            delegated_free=self.delegated_free.to_proto(),
            delegated_vesting=self.delegated_vesting.to_proto(),
            end_time=self.end_time,
        )



import copy
import json
from abc import ABC
from datetime import datetime
from typing import Any

from terra_sdk.util.converter import to_isoformat


def to_data(x: Any) -> Any:
    if hasattr(x, "to_data"):
        return x.to_data()
    if isinstance(x, int):
        return str(x)
    if isinstance(x, datetime):
        return to_isoformat(x)
    if isinstance(x, list):
        return [to_data(g) for g in x]
    if isinstance(x, dict):
        return dict_to_data(x)
    if isinstance(x, datetime):
        return to_isoformat(x)
    return x


def to_amino(x: Any) -> Any:
    if "to_amino" in dir(x):
        return x.to_amino()
    if isinstance(x, list):
        return [to_data(g) for g in x]
    if isinstance(x, datetime):
        return to_isoformat(x)
    if isinstance(x, dict):
        return dict_to_amino(x)
    if isinstance(x, int):
        return str(x)
    if isinstance(x, datetime):
        return to_isoformat(x)


def dict_to_amino(d: dict):
    return {key: to_amino(d[key]) for key in d}


def dict_to_data(d: dict) -> dict:
    """Recursively calls to_data on dict"""
    return {key: to_data(d[key]) for key in d}


class JSONSerializable(ABC):
    def to_data(self) -> Any:
        """Converts the object to its JSON-serializable Python data representation."""
        return dict_to_data(copy.deepcopy(self.__dict__))

    def to_json(self) -> str:
        """Marshals the object into a stringified JSON serialization. Keys are first sorted
        and the JSON rendered removes all unnecessary whitespace.

        Returns:
           str: JSON string representation
        """
        return json.dumps(self.to_data(), sort_keys=True, separators=(",", ":"))



import asyncio

import wrapt


class BaseAsyncAPI:
    # c = AsyncLCDClient
    def __init__(self, c):
        self._c = c

    def _run_sync(self, coroutine):
        """Runs an asynchronous coroutine synchronously."""
        return self._c.loop.run_until_complete(coroutine)

    @staticmethod
    async def _try_await(aw):
        """Checks if aw is a coroutine object and awaits it if so. Otherwise, just returns."""
        if asyncio.iscoroutine(aw):
            return await aw
        else:
            return aw


def sync_bind(async_call):
    """A decorator that redirects the function to the synchronous version of async_call."""

    @wrapt.decorator
    def decorator(wrapped, instance, args, kwargs):
        return instance._run_sync(async_call(instance, *args, **kwargs))

    return decorator



from __future__ import annotations

from asyncio import AbstractEventLoop, get_event_loop
from json import JSONDecodeError
from typing import List, Optional, Union

import nest_asyncio
from aiohttp import ClientSession
from multidict import CIMultiDict

from terra_sdk.core import Coins, Dec, Numeric
from terra_sdk.exceptions import LCDResponseError
from terra_sdk.key.key import Key
from terra_sdk.util.json import dict_to_data
from terra_sdk.util.url import urljoin

from .api.auth import AsyncAuthAPI, AuthAPI
from .api.authz import AsyncAuthzAPI, AuthzAPI
from .api.bank import AsyncBankAPI, BankAPI
from .api.distribution import AsyncDistributionAPI, DistributionAPI
from .api.feegrant import AsyncFeeGrantAPI, FeeGrantAPI
from .api.gov import AsyncGovAPI, GovAPI
from .api.ibc import AsyncIbcAPI, IbcAPI
from .api.ibc_transfer import AsyncIbcTransferAPI, IbcTransferAPI
from .api.mint import AsyncMintAPI, MintAPI
from .api.slashing import AsyncSlashingAPI, SlashingAPI
from .api.staking import AsyncStakingAPI, StakingAPI
from .api.tendermint import AsyncTendermintAPI, TendermintAPI
from .api.tx import AsyncTxAPI, TxAPI
from .api.wasm import AsyncWasmAPI, WasmAPI
from .lcdutils import AsyncLCDUtils, LCDUtils
from .params import APIParams
from .wallet import AsyncWallet, Wallet


def get_default(chain_id: str) -> [Coins, Numeric]:
    if chain_id == "phoenix-1":
        return [Coins.from_str("0.15uluna"), Numeric.parse(1.75)]
    if chain_id == "pisco-1":
        return [Coins.from_str("0.15uluna"), Numeric.parse(1.75)]
    if chain_id == "localterra":
        return [Coins.from_str("0.15uluna"), Numeric.parse(1.75)]

    raise ValueError("chain_id is invalid")


class AsyncLCDClient:
    def __init__(
        self,
        url: str,
        chain_id: Optional[str] = None,
        gas_prices: Optional[Coins.Input] = None,
        gas_adjustment: Optional[Numeric.Input] = None,
        loop: Optional[AbstractEventLoop] = None,
        _create_session: bool = True,  # don't create a session (used for sync LCDClient)
    ):
        if loop is None:
            loop = get_event_loop()
        self.loop = loop
        if _create_session:
            self.session = ClientSession(
                headers={"Accept": "application/json"}, loop=self.loop
            )

        self.chain_id = chain_id
        self.url = url
        self.last_request_height = None

        default_price, default_adjustment = get_default(chain_id)
        self.gas_prices = Coins(gas_prices) if gas_prices else default_price
        self.gas_adjustment = gas_adjustment if gas_adjustment else default_adjustment

        self.auth = AsyncAuthAPI(self)
        self.bank = AsyncBankAPI(self)
        self.distribution = AsyncDistributionAPI(self)
        self.feegrant = AsyncFeeGrantAPI(self)
        self.gov = AsyncGovAPI(self)
        self.mint = AsyncMintAPI(self)
        self.authz = AsyncAuthzAPI(self)
        self.slashing = AsyncSlashingAPI(self)
        self.staking = AsyncStakingAPI(self)
        self.tendermint = AsyncTendermintAPI(self)
        self.wasm = AsyncWasmAPI(self)
        self.ibc = AsyncIbcAPI(self)
        self.ibc_transfer = AsyncIbcTransferAPI(self)
        self.tx = AsyncTxAPI(self)
        self.utils = AsyncLCDUtils(self)

    def wallet(self, key: Key) -> AsyncWallet:
        """Creates a :class:`AsyncWallet` object from a key.

        Args:
            key (Key): key implementation
        """
        return AsyncWallet(self, key)

    async def _get(
        self,
        endpoint: str,
        params: Optional[Union[APIParams, CIMultiDict, list, dict]] = None,
        # raw: bool = False
    ):
        if (
            params
            and hasattr(params, "to_dict")
            and callable(getattr(params, "to_dict"))
        ):
            params = params.to_dict()

        async with self.session.get(
            urljoin(self.url, endpoint), params=params
        ) as response:
            try:
                result = await response.json(content_type=None)
            except JSONDecodeError:
                raise LCDResponseError(message=str(response.reason), response=response)
            if not 200 <= response.status < 299:
                raise LCDResponseError(message=str(result), response=response)
        self.last_request_height = (
            result.get("height") if result else self.last_request_height
        )
        return result  # if raw else result["result"]

    async def _post(
        self, endpoint: str, data: Optional[dict] = None  # , raw: bool = False
    ):
        async with self.session.post(
            urljoin(self.url, endpoint), json=data and dict_to_data(data)
        ) as response:
            try:
                result = await response.json(content_type=None)
            except JSONDecodeError:
                raise LCDResponseError(message=str(response.reason), response=response)
            if not 200 <= response.status < 299:
                raise LCDResponseError(message=result.get("message"), response=response)
        self.last_request_height = (
            result.get("height") if result else self.last_request_height
        )
        return result  # if raw else result["result"]

    async def _search(
        self,
        events: List[list],
        params: Optional[Union[APIParams, CIMultiDict, list, dict]] = None,
        # raw: bool = False
    ):

        actual_params : CIMultiDict = CIMultiDict()

        for event in events:
            if event[0] == "tx.height":
                actual_params.add("events", f"{event[0]}={event[1]}")
            else:
                actual_params.add("events", f"{event[0]}='{event[1]}'")
        if params:
            for p in params:
                actual_params.add(p, params[p])

        async with self.session.get(
            urljoin(self.url, "/cosmos/tx/v1beta1/txs"), params=actual_params
        ) as response:
            try:
                result = await response.json(content_type=None)
            except JSONDecodeError:
                raise LCDResponseError(message=str(response.reason), response=response)
            if not 200 <= response.status < 299:
                raise LCDResponseError(message=str(result), response=response)
        self.last_request_height = (
            result.get("height") if result else self.last_request_height
        )
        return result  # if raw else result["result"]

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.session.close()


class LCDClient(AsyncLCDClient):
    """An object representing a connection to a node running the Terra LCD server."""

    url: str
    """URL endpoint of LCD server."""

    chain_id: str
    """Chain ID of blockchain network connecting to."""

    gas_prices: Coins.Input
    """Gas prices to use for automatic fee estimation."""

    gas_adjustment: Union[str, float, int, Dec]
    """Gas adjustment factor for automatic fee estimation."""

    last_request_height: Optional[int]  # type: ignore
    """Height of response of last-made made LCD request."""

    auth: AuthAPI
    """:class:`AuthAPI<terra_sdk.client.lcd.api.auth.AuthAPI>`."""

    bank: BankAPI
    """:class:`BankAPI<terra_sdk.client.lcd.api.bank.BankAPI>`."""

    distribution: DistributionAPI
    """:class:`DistributionAPI<terra_sdk.client.lcd.api.distribution.DistributionAPI>`."""

    gov: GovAPI
    """:class:`GovAPI<terra_sdk.client.lcd.api.gov.GovAPI>`."""

    feegrant: FeeGrantAPI
    """:class:`FeeGrant<terra_sdk.client.lcd.api.feegrant.FeeGrantAPI>`."""

    mint: MintAPI
    """:class:`MintAPI<terra_sdk.client.lcd.api.mint.MintAPI>`."""

    authz: AuthzAPI
    """:class:`AuthzAPI<terra_sdk.client.lcd.api.authz.AuthzAPI>`."""

    slashing: SlashingAPI
    """:class:`SlashingAPI<terra_sdk.client.lcd.api.slashing.SlashingAPI>`."""

    staking: StakingAPI
    """:class:`StakingAPI<terra_sdk.client.lcd.api.staking.StakingAPI>`."""

    tendermint: TendermintAPI
    """:class:`TendermintAPI<terra_sdk.client.lcd.api.tendermint.TendermintAPI>`."""

    wasm: WasmAPI
    """:class:`WasmAPI<terra_sdk.client.lcd.api.wasm.WasmAPI>`."""

    tx: TxAPI
    """:class:`TxAPI<terra_sdk.client.lcd.api.tx.TxAPI>`."""

    ibc: IbcAPI
    """:class:`IbcAPI<terra_sdk.client.lcd.api.ibc.IbcAPI>`."""

    ibc_transfer: IbcTransferAPI
    """:class:`IbcTransferAPI<terra_sdk.client.lcd.api.ibc_transfer.IbcTransferAPI>`."""

    def __init__(
        self,
        url: str,
        chain_id: str = None,
        gas_prices: Optional[Coins.Input] = None,
        gas_adjustment: Optional[Numeric.Input] = None,
    ):
        super().__init__(
            url,
            chain_id,
            gas_prices,
            gas_adjustment,
            _create_session=False,
            loop=nest_asyncio.apply(get_event_loop()),
        )

        self.auth = AuthAPI(self)
        self.bank = BankAPI(self)
        self.distribution = DistributionAPI(self)
        self.gov = GovAPI(self)
        self.feegrant = FeeGrantAPI(self)
        self.mint = MintAPI(self)
        self.authz = AuthzAPI(self)
        self.slashing = SlashingAPI(self)
        self.staking = StakingAPI(self)
        self.tendermint = TendermintAPI(self)
        self.wasm = WasmAPI(self)
        self.ibc = IbcAPI(self)
        self.ibc_transfer = IbcTransferAPI(self)
        self.tx = TxAPI(self)
        self.utils = LCDUtils(self)

    async def __aenter__(self):
        raise NotImplementedError(
            "async context manager not implemented - you probably want AsyncLCDClient"
        )

    async def __aexit__(self, exc_type, exc, tb):
        raise NotImplementedError(
            "async context manager not implemented - you probably want AsyncLCDClient"
        )

    def wallet(self, key: Key) -> Wallet:  # type: ignore
        """Creates a :class:`Wallet` object from a key for easy transaction creating and
        signing.

        Args:
            key (Key): key implementation
        """
        return Wallet(self, key)

    async def _get(self, *args, **kwargs):
        # session has to be manually created and torn down for each HTTP request in a
        # synchronous client
        self.session = ClientSession(
            headers={"Accept": "application/json"}, loop=self.loop
        )
        try:
            result = await super()._get(*args, **kwargs)
        finally:
            await self.session.close()
        return result

    async def _post(self, *args, **kwargs):
        # session has to be manually created and torn down for each HTTP request in a
        # synchronous client
        self.session = ClientSession(
            headers={"Accept": "application/json"}, loop=self.loop
        )
        try:
            result = await super()._post(*args, **kwargs)
        finally:
            await self.session.close()
        return result

    async def _search(self, *args, **kwargs):
        # session has to be manually created and torn down for each HTTP request in a
        # synchronous client
        self.session = ClientSession(
            headers={"Accept": "application/json"}, loop=self.loop
        )
        try:
            result = await super()._search(*args, **kwargs)
        finally:
            await self.session.close()
        return result



from .msgs import (
    MsgClearAdmin,
    MsgExecuteContract,
    MsgInstantiateContract,
    MsgMigrateContract,
    MsgStoreCode,
    MsgUpdateAdmin,
)

__all__ = [
    "MsgStoreCode",
    "MsgInstantiateContract",
    "MsgExecuteContract",
    "MsgMigrateContract",
    "MsgUpdateAdmin",
    "MsgClearAdmin",
]



from __future__ import annotations

import copy
from datetime import datetime
from typing import Union

import attr
from dateutil import parser
from terra_proto.cosmos.staking.v1beta1 import BondStatus
from terra_proto.cosmos.staking.v1beta1 import Commission as Commission_pb
from terra_proto.cosmos.staking.v1beta1 import CommissionRates as CommissionRates_pb
from terra_proto.cosmos.staking.v1beta1 import Description as Description_pb
from terra_proto.cosmos.staking.v1beta1 import Validator as Validator_pb

from terra_sdk.core import Dec, ValAddress, ValConsPubKey
from terra_sdk.util.converter import to_isoformat
from terra_sdk.util.json import JSONSerializable

__all__ = ["CommissionRates", "Commission", "Description", "Validator", "BondStatus"]


def convert_bond_status_to_json(status: BondStatus) -> str:
    if status == BondStatus.BOND_STATUS_UNSPECIFIED:
        return "BOND_STATUS_UNSPECIFIED"
    elif status == BondStatus.BOND_STATUS_UNBONDED:
        return "BOND_STATUS_UNBONDED"
    elif status == BondStatus.BOND_STATUS_UNBONDING:
        return "BOND_STATUS_UNBONDING"
    elif status == BondStatus.BOND_STATUS_BONDED:
        return "BOND_STATUS_BONDED"


def convert_bond_status_from_json(status: str) -> BondStatus:
    if status == 0 or status == "BOND_STATUS_UNSPECIFIED":
        return BondStatus.BOND_STATUS_UNSPECIFIED
    elif status == 1 or status == "BOND_STATUS_UNBONDED":
        return BondStatus.BOND_STATUS_UNBONDED
    elif status == 2 or status == "BOND_STATUS_UNBONDING":
        return BondStatus.BOND_STATUS_UNBONDING
    elif status == 3 or status == "BOND_STATUS_BONDED":
        return BondStatus.BOND_STATUS_BONDED


@attr.s
class CommissionRates(JSONSerializable):
    """Data structure for validator's commission rates & policy."""

    rate: Dec = attr.ib(converter=Dec)
    """Current % commission rate."""

    max_rate: Dec = attr.ib(converter=Dec)
    """Maximum % commission rate permitted by policy."""

    max_change_rate: Dec = attr.ib(converter=Dec)
    """Maximum % change of commission per day."""

    def to_amino(self) -> dict:
        return {
            "rate": str(self.rate),
            "max_rate": str(self.max_rate),
            "max_change_rate": str(self.max_change_rate),
        }

    def to_data(self) -> dict:
        return {
            "rate": self.rate.to_data(),
            "max_rate": self.max_rate.to_data(),
            "max_change_rate": self.max_change_rate.to_data(),
        }

    @classmethod
    def from_data(cls, data: dict) -> CommissionRates:
        return cls(
            rate=data["rate"],
            max_rate=data["max_rate"],
            max_change_rate=data["max_change_rate"],
        )

    def to_proto(self) -> CommissionRates_pb:
        return CommissionRates_pb(
            rate=str(self.rate),
            max_rate=str(self.max_rate),
            max_change_rate=str(self.max_change_rate),
        )

    @classmethod
    def from_proto(cls, proto: CommissionRates_pb) -> CommissionRates:
        return cls(
            rate=proto.rate,
            max_rate=proto.max_rate,
            max_change_rate=proto.max_change_rate,
        )


@attr.s
class Commission(JSONSerializable):
    """Contains information about validator's commission rates."""

    commission_rates: CommissionRates = attr.ib()
    """Validator commission rates."""

    update_time: datetime = attr.ib(converter=parser.parse)
    """Last time commission rates were updated."""

    def to_amino(self) -> dict:
        return {
            "commission_rates": self.commission_rates.to_amino(),
            "update_time": to_isoformat(self.update_time),
        }

    def to_amino(self) -> dict:
        return {
            "commission_rates": self.commission_rates.to_data(),
            "update_time": to_isoformat(self.update_time),
        }

    @classmethod
    def from_data(cls, data: dict) -> Commission:
        return cls(
            commission_rates=CommissionRates.from_data(data["commission_rates"]),
            update_time=data["update_time"],
        )

    def to_proto(self) -> Commission_pb:
        return Commission_pb(
            commission_rates=self.commission_rates.to_proto(),
            update_time=self.update_time,
        )


@attr.s
class Description(JSONSerializable):
    """Validator description entry.

    Args:
        moniker: validator name, aka: \"Terran One\"
        identity: keybase.io identifier (used for setting logo)
        website: validator website
        details: longer description of validator
        security_contact: contact information for validator
    """

    DO_NOT_MODIFY = "[do-not-modify]"
    """"""

    moniker: str = attr.ib(default="")
    identity: str = attr.ib(default="")
    website: str = attr.ib(default="")
    details: str = attr.ib(default="")
    security_contact: str = attr.ib(default="")

    def to_amino(self) -> dict:
        return {
            "moniker": self.moniker,
            "identity": self.identity,
            "website": self.website,
            "details": self.details,
            "security_contact": self.security_contact,
        }

    def to_data(self) -> dict:
        return {
            "moniker": self.moniker,
            "identity": self.identity,
            "website": self.website,
            "details": self.details,
            "security_contact": self.security_contact,
        }

    @classmethod
    def from_data(cls, data) -> Description:
        return cls(
            data.get("moniker"),
            data.get("identity"),
            data.get("website"),
            data.get("details"),
            data.get("security_contact"),
        )

    def to_proto(self) -> Description_pb:
        return Description_pb(
            moniker=self.moniker,
            identity=self.identity,
            website=self.website,
            details=self.details,
            security_contact=self.security_contact,
        )

    @classmethod
    def from_proto(cls, proto: Description_pb) -> Description:
        return cls(
            proto.moniker,
            proto.identity,
            proto.website,
            proto.details,
            proto.security_contact,
        )


@attr.s
class Validator(JSONSerializable):
    """Contains information about a registered validator."""

    operator_address: ValAddress = attr.ib()
    """"""

    consensus_pubkey: ValConsPubKey = attr.ib()
    """"""

    jailed: bool = attr.ib(converter=bool)
    """"""

    status: BondStatus = attr.ib(converter=BondStatus)
    """"""

    tokens: int = attr.ib(converter=int)
    """"""

    delegator_shares: Dec = attr.ib(converter=Dec)
    """"""

    description: Description = attr.ib()
    """"""

    unbonding_height: int = attr.ib(converter=int)
    """"""

    unbonding_time: datetime = attr.ib(converter=parser.parse)
    """"""

    commission: Commission = attr.ib()
    """"""

    min_self_delegation: int = attr.ib(converter=int)
    """"""

    def to_amino(self) -> dict:
        return {
            "operator_address": self.operator_address,
            "consensus_pubkey": self.consensus_pubkey,
            "jailed": self.jailed,
            "status": self.status,
            "tokens": str(self.tokens),
            "delegator_shares": str(self.delegator_shares),
            "description": self.description.to_amino(),
            "unbonding_height": str(self.unbonding_height),
            "unbonding_time": to_isoformat(self.unbonding_time),
            "commission": self.commission.to_amino(),
            "min_self_delegation": str(self.min_self_delegation),
        }

    def to_data(self) -> dict:
        return {
            "operator_address": self.operator_address,
            "consensus_pubkey": self.consensus_pubkey,
            "jailed": self.jailed,
            "status": convert_bond_status_to_json(self.status),
            "tokens": str(self.tokens),
            "delegator_shares": str(self.delegator_shares),
            "description": self.description.to_amino(),
            "unbonding_height": str(self.unbonding_height),
            "unbonding_time": to_isoformat(self.unbonding_time),
            "commission": self.commission.to_amino(),
            "min_self_delegation": str(self.min_self_delegation),
        }

    @classmethod
    def from_data(cls, data: dict) -> Validator:
        return cls(
            operator_address=data["operator_address"],
            consensus_pubkey=data["consensus_pubkey"],
            jailed=data.get("jailed"),
            status=convert_bond_status_from_json(data["status"]),
            tokens=data["tokens"],
            delegator_shares=data["delegator_shares"],
            description=Description.from_data(data["description"]),
            unbonding_height=data.get("unbonding_height") or 0,
            unbonding_time=data["unbonding_time"],
            commission=Commission.from_data(data["commission"]),
            min_self_delegation=data["min_self_delegation"],
        )

    def to_proto(self) -> Validator_pb:
        return Validator_pb(
            operator_address=self.operator_address,
            consensus_pubkey=self.consensus_pubkey.to_proto(),
            jailed=self.jailed,
            status=self.status,
            tokens=str(self.tokens),
            delegator_shares=str(self.delegator_shares),
            description=self.description.to_proto(),
            unbonding_height=self.unbonding_height,
            unbonding_time=self.unbonding_time,
            commission=self.commission.to_proto(),
            min_self_delegation=str(self.min_self_delegation),
        )



import abc
import copy
from typing import Optional

import attr

from terra_sdk.core import (
    AccAddress,
    AccPubKey,
    ModeInfo,
    ModeInfoSingle,
    SignatureV2,
    SignDoc,
    ValAddress,
    ValPubKey,
)
from terra_sdk.core.bech32 import get_bech
from terra_sdk.core.public_key import (
    PublicKey,
    address_from_public_key,
    amino_pubkey_from_public_key,
)
from terra_sdk.core.signature_v2 import Descriptor
from terra_sdk.core.signature_v2 import Single as SingleDescriptor
from terra_sdk.core.tx import AuthInfo, SignerInfo, SignMode, Tx

__all__ = ["Key", "SignOptions"]


@attr.s
class SignOptions:
    account_number: int = attr.ib(converter=int)
    sequence: int = attr.ib(converter=int)
    sign_mode: SignMode = attr.ib()
    chain_id: str = attr.ib()


class Key:
    """Abstract Key interface, representing an agent with transaction-signing capabilities.

    Args:
        public_key (Optional[bytes]): compressed public key bytes,
    """

    public_key: Optional[PublicKey]
    """Compressed public key bytes, used to derive :data:`raw_address` and :data:`raw_pubkey`."""

    raw_address: Optional[bytes]
    """Raw Bech32 words of address, used to derive associated account and validator
    operator addresses.
    """

    raw_pubkey: Optional[bytes]
    """Raw Bech32 words of pubkey, used to derive associated account and validator
    pubkeys.
    """

    def __init__(self, public_key: Optional[PublicKey] = None):
        self.public_key = public_key
        if public_key:
            self.raw_address = address_from_public_key(public_key)
            self.raw_pubkey = amino_pubkey_from_public_key(public_key)

    @abc.abstractmethod
    def sign(self, payload: bytes) -> bytes:
        """Signs the data payload. An implementation of Key is expected to override this method.

        Args:
            payload (bytes): arbitrary data payload

        Raises:
            NotImplementedError: if not implemented

        Returns:
            bytes: signed payload
        """
        raise NotImplementedError("an instance of Key must implement Key.sign")

    @property
    def acc_address(self) -> AccAddress:
        """Terra Bech32 account address. Default derivation via :data:`public_key` is provided.

        Raises:
            ValueError: if Key was not initialized with proper public key

        Returns:
            AccAddress: account address
        """
        if not self.raw_address:
            raise ValueError("could not compute acc_address: missing raw_address")
        return AccAddress(get_bech("terra", self.raw_address.hex()))

    @property
    def val_address(self) -> ValAddress:
        """Terra Bech32 validator operator address. Default derivation via :data:`public_key` is provided.

        Raises:
            ValueError: if Key was not initialized with proper public key

        Returns:
            ValAddress: validator operator address
        """
        if not self.raw_address:
            raise ValueError("could not compute val_address: missing raw_address")
        return ValAddress(get_bech("terravaloper", self.raw_address.hex()))

    @property
    def acc_pubkey(self) -> AccPubKey:
        """Terra Bech32 account pubkey. Default derivation via :data:`public_key` is provided.

        Raises:
            ValueError: if Key was not initialized with proper public key

        Returns:
            AccPubKey: account pubkey
        """
        if not self.raw_pubkey:
            raise ValueError("could not compute acc_pubkey: missing raw_pubkey")
        return AccPubKey(get_bech("terrapub", self.raw_pubkey.hex()))

    @property
    def val_pubkey(self) -> ValPubKey:
        """Terra Bech32 validator pubkey. Default derivation via ``public_key`` is provided.

        Raises:
            ValueError: if Key was not initialized with proper public key

        Returns:
            ValPubKey: validator pubkey
        """
        if not self.raw_pubkey:
            raise ValueError("could not compute val_pubkey: missing raw_pubkey")
        return ValPubKey(get_bech("terravaloperpub", self.raw_pubkey.hex()))

    def create_signature_amino(self, sign_doc: SignDoc) -> SignatureV2:
        if self.public_key is None:
            raise ValueError(
                "signature could not be created: Key instance missing public_key"
            )

        return SignatureV2(
            public_key=self.public_key,
            data=Descriptor(
                SingleDescriptor(
                    mode=SignMode.SIGN_MODE_LEGACY_AMINO_JSON,
                    signature=(self.sign(sign_doc.to_amino_json()))
                    # signature=base64.b64encode(self.sign(sign_doc.to_amino_json()))
                )
            ),
            sequence=sign_doc.sequence,
        )

    def create_signature(self, sign_doc: SignDoc) -> SignatureV2:
        """Signs the transaction with the signing algorithm provided by this Key implementation,
        and outputs the signature. The signature is only returned, and must be manually added to
        the ``signatures`` field of an :class:`Tx`.

        Args:
            sign_doc (SignDoc): unsigned transaction

        Raises:
            ValueError: if missing ``public_key``

        Returns:
            SignatureV2: signature object
        """
        if self.public_key is None:
            raise ValueError(
                "signature could not be created: Key instance missing public_key"
            )

        # make backup
        si_backup = copy.deepcopy(sign_doc.auth_info.signer_infos)
        sign_doc.auth_info.signer_infos = [
            SignerInfo(
                public_key=self.public_key,
                sequence=sign_doc.sequence,
                mode_info=ModeInfo(
                    single=ModeInfoSingle(mode=SignMode.SIGN_MODE_DIRECT)
                ),
            )
        ]
        signature = self.sign(sign_doc.to_bytes())

        # restore
        sign_doc.auth_info.signer_infos = si_backup

        return SignatureV2(
            public_key=self.public_key,
            data=Descriptor(
                single=SingleDescriptor(
                    mode=SignMode.SIGN_MODE_DIRECT, signature=signature
                )
            ),
            sequence=sign_doc.sequence,
        )

    def sign_tx(self, tx: Tx, options: SignOptions) -> Tx:
        """Signs the transaction with the signing algorithm provided by this Key implementation,
        and creates a ready-to-broadcast :class:`Tx` object with the signature applied.

        Args:
            tx (Tx): unsigned transaction
            options (SignOptions): options for signing

        Returns:
            Tx: ready-to-broadcast transaction object
        """
        signedTx = Tx(
            body=tx.body,
            auth_info=AuthInfo(signer_infos=[], fee=tx.auth_info.fee),
            signatures=[],
        )
        signDoc = SignDoc(
            chain_id=options.chain_id,
            account_number=options.account_number,
            sequence=options.sequence,
            auth_info=signedTx.auth_info,
            tx_body=signedTx.body,
        )

        if options.sign_mode == SignMode.SIGN_MODE_LEGACY_AMINO_JSON:
            signature: SignatureV2 = self.create_signature_amino(signDoc)
        else:
            signature: SignatureV2 = self.create_signature(signDoc)

        sigData: SingleDescriptor = signature.data.single
        for sig in tx.signatures:
            signedTx.signatures.append(sig)
        signedTx.signatures.append(sigData.signature)
        for infos in tx.auth_info.signer_infos:
            signedTx.auth_info.signer_infos.append(infos)
        signedTx.auth_info.signer_infos.append(
            SignerInfo(
                public_key=signature.public_key,
                sequence=signature.sequence,
                mode_info=ModeInfo(single=ModeInfoSingle(mode=sigData.mode)),
            )
        )
        return signedTx



"""Slashing module messages types."""

from __future__ import annotations

import attr
from terra_proto.cosmos.slashing.v1beta1 import MsgUnjail as MsgUnjail_pb

from terra_sdk.core import ValAddress
from terra_sdk.core.msg import Msg

__all__ = ["MsgUnjail"]


@attr.s
class MsgUnjail(Msg):
    """Attempt to unjail a jailed validator (must be submitted by same validator).

    Args:
        address: validator address to unjail"""

    type_amino = "cosmos-sdk/MsgUnjail"
    """"""
    type_url = "/cosmos.slashing.v1beta1.MsgUnjail"
    """"""
    action = "unjail"
    """"""
    prototype = MsgUnjail_pb
    """"""

    address: ValAddress = attr.ib()

    def to_amino(self) -> dict:
        return {"type": self.type_amino, "value": {"address": self.address}}

    @classmethod
    def from_data(cls, data: dict) -> MsgUnjail:
        return cls(address=data["address"])

    def to_proto(self) -> MsgUnjail_pb:
        return MsgUnjail_pb(validator_addr=self.address)

    @classmethod
    def from_proto(cls, proto: MsgUnjail_pb) -> MsgUnjail:
        return cls(address=proto.address)



from typing import Dict

import attr

from terra_sdk.core import AccAddress, Coins, ValAddress

from ._base import BaseAsyncAPI, sync_bind

__all__ = ["AsyncDistributionAPI", "DistributionAPI", "Rewards"]


@attr.s
class Rewards:
    rewards: Dict[ValAddress, Coins] = attr.ib()
    """Delegator rewards, indexed by validator operator address."""

    total: Coins = attr.ib()
    """Total sum of rewards."""


class AsyncDistributionAPI(BaseAsyncAPI):
    async def rewards(self, delegator: AccAddress) -> Rewards:
        """Fetches the staking reward data for a delegator.

        Args:
            delegator (AccAddress): delegator account address

        Returns:
            Rewards: delegator rewards
        """
        res = await self._c._get(
            f"/cosmos/distribution/v1beta1/delegators/{delegator}/rewards"
        )
        return Rewards(
            rewards={
                item["validator_address"]: Coins.from_data(item["reward"] or [])
                for item in res["rewards"]
            },
            total=Coins.from_data(res["total"]),
        )

    async def validator_commission(self, validator: ValAddress) -> Coins:
        """Fetches the commission reward data for a validator.

        Args:
            validator (ValAddress): validator operator address

        Returns:
            ValidatorCommission: validator rewards
        """
        res = await self._c._get(
            f"/cosmos/distribution/v1beta1/validators/{validator}/commission"
        )
        commission = res["commission"]
        return Coins.from_data(commission["commission"])

    async def withdraw_address(self, delegator: AccAddress) -> AccAddress:
        """Fetches the withdraw address associated with a delegator.

        Args:
            delegator (AccAddress): delegator account address

        Returns:
            AccAddress: withdraw address
        """
        res = await self._c._get(
            f"/cosmos/distribution/v1beta1/delegators/{delegator}/withdraw_address"
        )
        return res.get("withdraw_address")

    async def community_pool(self) -> Coins:
        """Fetches the community pool.

        Returns:
            Coins: community pool
        """
        res = await self._c._get("/cosmos/distribution/v1beta1/community_pool")
        return Coins.from_data(res.get("pool"))

    async def parameters(self) -> dict:
        """Fetches the Distribution module parameters.

        Returns:
            dict: Distribution module parameters
        """
        res = await self._c._get("/cosmos/distribution/v1beta1/params")
        return res.get("params")


class DistributionAPI(AsyncDistributionAPI):
    @sync_bind(AsyncDistributionAPI.rewards)
    def rewards(self, delegator: AccAddress) -> Rewards:
        pass

    rewards.__doc__ = AsyncDistributionAPI.rewards.__doc__

    @sync_bind(AsyncDistributionAPI.validator_commission)
    def validator_commission(self, validator: ValAddress) -> Coins:
        pass

    validator_commission.__doc__ = AsyncDistributionAPI.validator_commission.__doc__

    @sync_bind(AsyncDistributionAPI.withdraw_address)
    def withdraw_address(self, delegator: AccAddress) -> AccAddress:
        pass

    withdraw_address.__doc__ = AsyncDistributionAPI.withdraw_address.__doc__

    @sync_bind(AsyncDistributionAPI.community_pool)
    def community_pool(self) -> Coins:
        pass

    community_pool.__doc__ = AsyncDistributionAPI.community_pool.__doc__

    @sync_bind(AsyncDistributionAPI.parameters)
    def parameters(self) -> dict:
        pass

    parameters.__doc__ = AsyncDistributionAPI.parameters.__doc__



from .data import (
    Authorization,
    AuthorizationGrant,
    GenericAuthorization,
    SendAuthorization,
    StakeAuthorization,
)
from .msgs import MsgExecAuthorized, MsgGrantAuthorization, MsgRevokeAuthorization

__all__ = [
    "MsgExecAuthorized",
    "MsgGrantAuthorization",
    "MsgRevokeAuthorization",
    "Authorization",
    "SendAuthorization",
    "GenericAuthorization",
    "StakeAuthorization",
    "AuthorizationGrant",
]



"""Gov module data types."""

from __future__ import annotations

from datetime import datetime
from typing import List

import attr
from dateutil import parser
from terra_proto.cosmos.gov.v1beta1 import Proposal as Proposal_pb
from terra_proto.cosmos.gov.v1beta1 import ProposalStatus
from terra_proto.cosmos.gov.v1beta1 import TallyResult as TallyResult_pb
from terra_proto.cosmos.gov.v1beta1 import Vote as Vote_pb
from terra_proto.cosmos.gov.v1beta1 import VoteOption
from terra_proto.cosmos.gov.v1beta1 import WeightedVoteOption as WeightedVoteOption_pb

from terra_sdk.core import AccAddress, Coins
from terra_sdk.util.converter import to_isoformat
from terra_sdk.util.json import JSONSerializable
from terra_sdk.util.parse_content import Content, parse_content

__all__ = ["Proposal", "Content", "VoteOption", "WeightedVoteOption", "ProposalStatus"]


@attr.s
class TallyResult(JSONSerializable):
    yes: str = attr.ib()
    abstain: str = attr.ib()
    no: str = attr.ib()
    no_with_veto: str = attr.ib()

    def to_amino(self) -> dict:
        return {
            "yes": self.yes,
            "abstain": self.abstain,
            "no": self.no,
            "no_with_veto": self.no_with_veto,
        }

    def to_data(self) -> dict:
        return {
            "yes": self.yes,
            "abstain": self.abstain,
            "no": self.no,
            "no_with_veto": self.no_with_veto,
        }

    @classmethod
    def from_data(cls, data: dict) -> TallyResult:
        return cls(
            yes=data["yes"],
            abstain=data["abstain"],
            no=data["no"],
            no_with_veto=data["no_with_veto"],
        )

    def to_proto(self) -> TallyResult_pb:
        return TallyResult_pb(
            yes=self.yes,
            abstain=self.abstain,
            no=self.no,
            no_with_veto=self.no_with_veto,
        )


@attr.s
class Proposal(JSONSerializable):
    """Contains information about a submitted proposal on the blockchain."""

    proposal_id: int = attr.ib(converter=int)
    """Proposal's ID."""

    content: Content = attr.ib()
    """Proposal contents."""

    status: str = attr.ib()
    """Status of proposal."""

    final_tally_result: TallyResult = attr.ib()
    """Final tallied result of the proposal (after vote)."""

    submit_time: datetime = attr.ib(converter=parser.parse)
    """Timestamp at which proposal was submitted."""

    deposit_end_time: datetime = attr.ib(converter=parser.parse)
    """Time at which the deposit period ended, or will end."""

    total_deposit: Coins = attr.ib(converter=Coins)
    """Total amount deposited for proposal"""

    voting_start_time: datetime = attr.ib(converter=parser.parse)
    """Time at which voting period started, or will start."""

    voting_end_time: datetime = attr.ib(converter=parser.parse)
    """Time at which voting period ended, or will end."""

    def to_amino(self) -> dict:
        return {
            "proposal_id": str(self.proposal_id),
            "content": self.content.to_amino(),
            "status": self.status,
            "final_tally_result": self.final_tally_result.to_amino(),
            "submit_time": to_isoformat(self.submit_time),
            "deposit_end_time": to_isoformat(self.deposit_end_time),
            "total_deposit": self.total_deposit.to_amino(),
            "voting_start_time": to_isoformat(self.voting_start_time),
            "voting_end_time": to_isoformat(self.voting_end_time),
        }

    @classmethod
    def from_data(cls, data: dict) -> Proposal:
        return cls(
            proposal_id=data["proposal_id"],
            content=parse_content(data["content"]),
            status=data["status"],
            final_tally_result=data["final_tally_result"],
            submit_time=data["submit_time"],
            deposit_end_time=data["deposit_end_time"],
            total_deposit=Coins.from_data(data["total_deposit"]),
            voting_start_time=data["voting_start_time"],
            voting_end_time=data["voting_end_time"],
        )

    def to_proto(self) -> Proposal_pb:
        return Proposal_pb(
            proposal_id=self.proposal_id,
            content=self.content.pack_any(),
            status=ProposalStatus.from_str(self.status),
            final_tally_result=self.final_tally_result.to_proto(),
            submit_time=self.submit_time,
            deposit_end_time=self.deposit_end_time,
            total_deposit=self.total_deposit.to_proto(),
            voting_start_time=self.voting_start_time,
            voting_end_time=self.voting_end_time,
        )


@attr.s
class WeightedVoteOption(JSONSerializable):
    weight: str = attr.ib()
    option: VoteOption = attr.ib(converter=int)

    def to_amino(self) -> dict:
        return {"weight": self.weight, "option": self.option.name}

    def to_data(self) -> dict:
        return {"weight": self.weight, "option": self.option.name}

    @classmethod
    def from_data(cls, data: dict) -> WeightedVoteOption:
        return cls(option=data["option"], weight=data["weight"])

    def to_proto(self) -> WeightedVoteOption_pb:
        return WeightedVoteOption_pb(option=self.option, weight=self.weight)


@attr.s
class Vote(JSONSerializable):
    proposal_id: int = attr.ib(converter=int)
    voter: AccAddress = attr.ib()
    options: List[WeightedVoteOption] = attr.ib(converter=list)

    def to_amino(self) -> dict:
        return {
            "proposal_id": str(self.proposal_id),
            "voter": self.voter,
            "options": [opt.to_amino() for opt in self.options],
        }

    def to_data(self) -> dict:
        return {
            "proposal_id": str(self.proposal_id),
            "voter": self.voter,
            "options": [opt.to_data() for opt in self.options],
        }

    @classmethod
    def from_data(cls, data: dict) -> Vote:
        return cls(
            proposal_id=data["proposal_id"],
            voter=data["voter"],
            options=data["options"],
        )

    def to_proto(self) -> Vote_pb:
        return Vote_pb(
            proposal_id=self.proposal_id, voter=self.voter, options=self.options
        )



"""Terra SDK-specific errors and exceptions."""


class LCDResponseError(IOError):
    """Triggered when response from LCD is not 2xx status code"""

    def __init__(self, message, response):
        self.message = message
        self.response = response

    def __str__(self):
        message = ""
        if self.message:
            message = " - " + self.message
        return f"Status {self.response.status}{message}"



"""Transaction broadcast result data types."""

from __future__ import annotations

from typing import List, Optional, Union

import attr

from terra_sdk.core.auth import TxLog, parse_tx_logs
from terra_sdk.util.json import JSONSerializable

__all__ = [
    "BlockTxBroadcastResult",
    "SyncTxBroadcastResult",
    "AsyncTxBroadcastResult",
    "is_tx_error",
]


@attr.s
class BlockTxBroadcastResult(JSONSerializable):
    """Data object that contains the response result from node after transaction
    has been broadcasted with the ``block`` broadcast mode."""

    height: int = attr.ib(converter=int)
    """Height at which transaction was included."""
    txhash: str = attr.ib()
    """Transaction hash."""
    raw_log: Optional[str] = attr.ib()
    """Raw JSON of transaction events."""
    gas_wanted: int = attr.ib(converter=int)
    """Gas requested by the transaction."""
    gas_used: int = attr.ib(converter=int)
    """Actual amount of gas consumed by transaction."""
    logs: Optional[List[TxLog]] = attr.ib(converter=parse_tx_logs)  # type: ignore
    """List of transaction logs."""
    code: Optional[int] = attr.ib(default=None)
    """If this is present, the transaction failed."""
    codespace: Optional[str] = attr.ib(default=None)
    """Error subspace name: used alongside ``code``."""
    info: Optional[str] = attr.ib(default=None)
    """"""
    data: Optional[str] = attr.ib(default=None)
    """"""
    timestamp: Optional[str] = attr.ib(default=None)
    """timestamp"""

    def is_tx_error(self) -> bool:
        """Returns whether the transaction failed."""
        return is_tx_error(self)


@attr.s
class SyncTxBroadcastResult(JSONSerializable):
    """Data object that contains the response result from node after transactionco
    has been broadcasted with the ``sync`` broadcast mode."""

    txhash: str = attr.ib()
    """Transaction hash."""
    raw_log: Optional[str] = attr.ib()
    """Raw JSON of transaction events."""
    code: Optional[int] = attr.ib(default=None)
    """If this is present, the transaction failed."""
    codespace: Optional[str] = attr.ib(default=None)
    """Error subspace name: used alongside ``code``."""

    def is_tx_error(self) -> bool:
        """Returns whether the transaction failed."""
        return is_tx_error(self)


@attr.s
class AsyncTxBroadcastResult(JSONSerializable):
    """Data object that contains the response result from node after transaction
    has been broadcasted with the ``sync`` broadcast mode."""

    txhash: str = attr.ib()
    """Transaction hash."""


def is_tx_error(result: Union[BlockTxBroadcastResult, SyncTxBroadcastResult]):
    """Returns whether the transaction failed."""
    return result.code != 0



"""Numeric types."""

from __future__ import annotations

import re
from decimal import Decimal
from typing import Union

from terra_sdk.util.json import JSONSerializable

DEC_NUM_DIGITS = 18
"""Number of digits for Decimal."""

DEC_ONE = 10**DEC_NUM_DIGITS
DEC_PATTERN = re.compile(r"^(\-)?(\d+)(\.(\d+))?\Z")

__all__ = ["DEC_NUM_DIGITS", "Dec", "Numeric"]


def convert_to_dec_bignum(arg: Union[str, int, float, Decimal]):
    if isinstance(arg, int) or isinstance(arg, Decimal):
        return int(arg * DEC_ONE)
    if isinstance(arg, float):
        arg = str("%f" % arg)
    if isinstance(arg, str):
        parts = DEC_PATTERN.match(arg)
        if parts is None:
            raise ValueError(f"Unable to parse Dec from string: {arg}")
        result = int(parts.group(2)) * DEC_ONE  # whole part
        if parts.group(3):
            fraction = int(parts.group(4)[0:DEC_NUM_DIGITS].ljust(DEC_NUM_DIGITS, "0"))
            result += fraction
        if parts.group(1):
            result *= -1
        return result
    else:
        raise TypeError(
            f"Unable to parse Dec integer representation from given argument {arg}"
        )


def chop_precision_and_round(d: int) -> int:
    """Cosmos-SDK's banker's rounding:
    https://github.com/cosmos/cosmos-sdk/blob/1d75e0e984e7132efd54c3526e36b3585e2d91c0/types/decimal.go#L491
    """
    if d < 0:
        return -1 * chop_precision_and_round(d * -1)

    quo, rem = d // DEC_ONE, d % DEC_ONE

    if rem == 0:
        return quo

    if rem < DEC_ONE / 2:
        return quo
    elif rem > DEC_ONE / 2:
        return quo + 1
    else:
        if quo % 2 == 0:
            return quo
        return quo + 1


class Dec(JSONSerializable):
    """BigInt-based Decimal representation with basic arithmetic operations with
    compatible Python numeric types (int, float, Decimal). Does not work with
    ``NaN``, ``Infinity``, ``+0``, ``-0``, etc. Serializes as a string with 18 points of
    decimal precision.

    >>> Dec(5)
    Dec("5.0")
    >>> Dec("121.1232")
    Dec("121.1232")
    >>> Dec(121.1232)
    Dec("121.1232")

    Args:
        arg (Union[str, int, float, Decimal, Dec]): argument to coerce into Dec
    """

    _i: int = 0

    def __init__(self, arg: Union[str, int, float, Decimal, Dec]):
        if isinstance(arg, Dec):
            self._i = arg._i
            return
        else:
            self._i = int(convert_to_dec_bignum(arg))

    @classmethod
    def zero(cls) -> Dec:
        """Dec representation of zero.

        Returns:
            Dec: zero
        """
        return cls(0)

    @classmethod
    def one(cls):
        """Dec representation of one.

        Returns:
            Dec: one
        """
        nd = cls(0)
        nd._i = DEC_ONE
        return nd

    def __str__(self) -> str:
        """Converts to a string using all 18 decimal precision points.

        Returns:
            str: string representation
        """
        if self._i == 0:
            return "0." + DEC_NUM_DIGITS * "0"
        parity = "-" if self._i < 0 else ""
        return f"{parity}{self.whole}.{self.frac}"

    def to_short_str(self) -> str:
        """Converts to a string, but truncates all unnecessary zeros.

        Returns:
            str: string representation
        """
        parity = "-" if self._i < 0 else ""
        frac = self.frac.rstrip("0")
        dot = "." if len(frac) > 0 else ""
        return f"{parity}{self.whole}{dot}{frac}"

    def __repr__(self):
        return f"Dec('{self.to_short_str()}')"  # short representation

    def __int__(self) -> int:
        int_part = abs(self._i) // DEC_ONE
        int_part *= -1 if self._i < 0 else 1
        return int_part

    def __float__(self) -> float:
        # NOTE: This is not robust enough for: float(Dec(float)) to give the same output
        # and should mainly be used as getting a rough value from the Dec object.
        return float(self._i) / DEC_ONE

    @property
    def parity(self) -> int:
        """Get the parity of the Dec value. Returns -1 if value is below 0, and 1 otherwise.

        Returns:
            int: parity
        """
        return -1 if self._i < 0 else 1

    @property
    def whole(self) -> str:
        """Get the integral part of the Dec value.

        Returns:
            str: integer, as string
        """
        return str(abs(self._i) // DEC_ONE)

    @property
    def frac(self) -> str:
        """Get the fractional part of the Dec value.

        Returns:
            str: fraction, as string
        """
        return str(abs(self._i) % DEC_ONE).rjust(DEC_NUM_DIGITS, "0")

    def to_data(self) -> str:
        return str(self)

    def __eq__(self, other) -> bool:
        if isinstance(other, str):
            return False
        else:
            return self._i == Dec(other)._i

    def lt(self, other: Union[str, int, float, Decimal, Dec]) -> bool:
        """Check less than.

        Args:
            other (Union[str, int, float, Decimal, Dec]): compared object
        """
        if isinstance(other, Dec):
            return self._i < other._i
        return (Decimal(self._i) / DEC_ONE) < other

    def __lt__(self, other: Union[str, int, float, Decimal, Dec]) -> bool:
        return self.lt(other)

    def le(self, other: Union[str, int, float, Decimal, Dec]) -> bool:
        """Check less than or equal to.

        Args:
            other (Union[str, int, float, Decimal, Dec]): compared object
        """
        return self < other or self.__eq__(other)

    def __le__(self, other: Union[str, int, float, Decimal, Dec]) -> bool:
        return self.le(other)

    def gt(self, other: Union[str, int, float, Decimal, Dec]) -> bool:
        """Check greater than.

        Args:
            other (Union[str, int, float, Decimal, Dec]): compared object
        """
        if isinstance(other, Dec):
            return self._i > other._i
        return (Decimal(self._i) / DEC_ONE) > other

    def __gt__(self, other) -> bool:
        return self.gt(other)

    def ge(self, other) -> bool:
        """Check greater than or equal to.

        Args:
            other (Union[str, int, float, Decimal, Dec]): compared object
        """
        return self.gt(other) or self.__eq__(other)

    def __ge__(self, other) -> bool:
        return self.ge(other)

    def add(self, addend: Union[str, int, float, Decimal, Dec]) -> Dec:
        """Performs addition. ``addend`` is first converted into Dec.

        Args:
            addend (Union[str, int, float, Decimal, Dec]): addend

        Returns:
            Dec: sum
        """
        nd = Dec.zero()
        nd._i = self._i + Dec(addend)._i
        return nd

    def __add__(self, addend: Union[str, int, float, Decimal, Dec]) -> Dec:
        return self.add(addend)

    def __radd__(self, addend: Union[str, int, float, Decimal, Dec]):
        return Dec(addend).add(self)

    def sub(self, subtrahend: Union[str, int, float, Decimal, Dec]) -> Dec:
        """Performs subtraction. ``subtrahend`` is first converted into Dec.

        Args:
            subtrahend (Union[str, int, float, Decimal, Dec]): subtrahend

        Returns:
            Dec: difference
        """
        nd = Dec.zero()
        nd._i = self._i - Dec(subtrahend)._i
        return nd

    def __sub__(self, subtrahend: Union[str, int, float, Decimal, Dec]) -> Dec:
        return self.sub(subtrahend)

    def __rsub__(self, minuend: Dec) -> Dec:
        return Dec(minuend).sub(self)

    def mul(self, multiplier: Union[str, int, float, Decimal, Dec]) -> Dec:
        """Performs multiplication. ``multiplier`` is first converted into Dec.

        Args:
            multiplier (Union[str, int, float, Decimal, Dec]): multiplier

        Returns:
            Dec: product
        """
        x = self._i
        y = Dec(multiplier)._i
        nd = Dec.zero()
        nd._i = chop_precision_and_round(x * y)
        return nd

    def __mul__(self, multiplier: Union[str, int, float, Decimal, Dec]) -> Dec:
        return self.mul(multiplier)

    def __rmul__(self, multiplicand: Union[str, int, float, Decimal, Dec]):
        return Dec(multiplicand).mul(self)

    def div(self, divisor: Union[str, int, float, Decimal, Dec]) -> Dec:
        """Performs division. ``divisor`` is first converted into Dec.
        It works like truediv('/')

        Args:
            divisor (Union[str, int, float, Decimal, Dec]): divisor

        Raises:
            ZeroDivisionError: if ``divisor`` is 0

        Returns:
            Dec: quotient
        """
        if Dec(divisor)._i == 0:
            raise ZeroDivisionError(f"tried to divide by 0: {self!r} / {divisor!r}")
        nd = Dec.zero()
        nd._i = chop_precision_and_round(self._i * DEC_ONE * DEC_ONE // Dec(divisor)._i)
        return nd

    def __truediv__(self, divisor) -> Dec:
        return self.div(divisor)

    def __rtruediv__(self, divisor) -> Dec:
        return Dec(divisor).div(self)

    def __floordiv__(self, divisor):
        return Dec(chop_precision_and_round(self.div(divisor).sub(0.5)._i))

    def __rfloordiv__(self, divisor):
        return Dec(chop_precision_and_round(divisor / self.sub(0.5)._i))

    def mod(self, modulo: Union[str, int, float, Decimal, Dec]) -> Dec:
        """Performs modulus. ``modulo`` is first converted into Dec.

        Args:
            modulo (Union[str, int, float, Decimal, Dec]): modulo

        Returns:
            Dec: modulus
        """
        return self.sub(self.__floordiv__(modulo).mul(modulo))

    def __mod__(self, modulo) -> Dec:
        return self.mod(modulo)

    def __neg__(self) -> Dec:
        x = Dec(self)
        x._i *= -1
        return x

    def __abs__(self) -> Dec:
        x = Dec(self)
        x._i = abs(x._i)
        return x

    def __pos__(self) -> Dec:
        # __pos__ implies a copy
        return Dec(self)

    @classmethod
    def from_data(cls, data: str) -> Dec:
        """Converts Dec-formatted string into proper :class:`Dec` object."""
        return cls(data)

    @classmethod
    def with_prec(cls, i: Union[int, str], prec: int) -> Dec:
        """Replicates Cosmos SDK's ``Dec.withPreic(i, prec)``.

        Args:
            i (Union[int, str]): numeric value
            prec (int): precision

        Returns:
            Dec: decimal
        """
        d = cls(0)
        i = int(i)
        d._i = i * 10 ** (DEC_NUM_DIGITS - int(prec))
        return d


class Numeric:

    Input = Union[str, int, float, Decimal, Dec]
    """"""

    Output = Union[int, Dec]
    """"""

    @staticmethod
    def parse(value: Numeric.Input) -> Numeric.Output:
        """Parses the value and coerces it into an ``int`` or :class:`Dec`.

        Args:
            value (Numeric.Input): value to be parsed

        Raises:
            TypeError: if supplied value could not be parsed

        Returns:
            Numeric.Output: coerced number
        """
        if isinstance(value, int) or isinstance(value, Dec):
            return value
        elif isinstance(value, str):
            if "." in value:
                return Dec(value)
            else:
                return int(value)
        elif isinstance(value, float) or isinstance(value, Decimal):
            return Dec(value)
        else:
            raise TypeError(f"could not parse numeric value to Dec or int: {value}")



"""Bank module message types."""

from __future__ import annotations

from typing import List

from betterproto.lib.google.protobuf import Any as Any_pb
from terra_proto.cosmos.bank.v1beta1 import Input as Input_pb
from terra_proto.cosmos.bank.v1beta1 import MsgMultiSend as MsgMultiSend_pb
from terra_proto.cosmos.bank.v1beta1 import MsgSend as MsgSend_pb
from terra_proto.cosmos.bank.v1beta1 import Output as Output_pb

from terra_sdk.core import AccAddress, Coins
from terra_sdk.core.msg import Msg
from terra_sdk.util.json import JSONSerializable

__all__ = ["MsgSend", "MsgMultiSend", "MultiSendInput", "MultiSendOutput"]

import attr


@attr.s
class MsgSend(Msg):
    """Sends native Terra assets (Luna or Terra stablecoins) from ``from_address`` to ``to_address``.

    Args:
        from_address (AccAddress): sender
        to_address (AccAddress): recipient
        amount (Coins): coins to send
    """

    type_amino = "cosmos-sdk/MsgSend"
    """"""
    type_url = "/cosmos.bank.v1beta1.MsgSend"
    """"""
    action = "send"
    """"""
    prototype = MsgSend_pb
    """"""

    from_address: AccAddress = attr.ib()
    to_address: AccAddress = attr.ib()
    amount: Coins = attr.ib(converter=Coins)

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "from_address": self.from_address,
                "to_address": self.to_address,
                "amount": self.amount.to_amino(),
            },
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgSend:
        return cls(
            from_address=data["from_address"],
            to_address=data["to_address"],
            amount=Coins.from_data(data["amount"]),
        )

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "from_address": self.from_address,
            "to_address": self.to_address,
            "amount": self.amount.to_data(),
        }

    @classmethod
    def from_proto(cls, proto: MsgSend_pb) -> MsgSend:
        return cls(
            from_address=proto.from_address,
            to_address=proto.to_address,
            amount=Coins.from_proto(proto.amount),
        )

    def to_proto(self) -> MsgSend_pb:
        proto = MsgSend_pb()
        proto.from_address = self.from_address
        proto.to_address = self.to_address
        proto.amount = [c.to_proto() for c in self.amount]
        return proto


@attr.s
class MultiSendInput(JSONSerializable):
    """Organizes data for MsgMultiSend input/outputs. Expects data to be provided in the
    format:

    Args:
         address (AccAddress): from_address
         coins (Coins): amount to send from the address
    """

    address: AccAddress = attr.ib()
    """Input / output address."""

    coins: Coins = attr.ib(converter=Coins)
    """Coins to be sent."""

    def to_amino(self) -> dict:
        return {"address": self.address, "coins": self.coins.to_amino()}

    def to_data(self) -> dict:
        return {"address": self.address, "coins": self.coins.to_data()}

    @classmethod
    def from_data(cls, data: dict):
        return cls(address=data["address"], coins=Coins.from_data(data["coins"]))

    @classmethod
    def from_proto(cls, proto: Input_pb) -> MultiSendInput:
        return cls(address=proto.address, coins=Coins.from_proto(proto.coins))

    def to_proto(self) -> Input_pb:
        return Input_pb(address=self.address, coins=self.coins.to_proto())


@attr.s
class MultiSendOutput(JSONSerializable):
    """Organizes data for MsgMultiSend input/outputs. Expects data to be provided in the
    format:

    Args:
         address (AccAddress): to_address
         coins (Coins): amount to receive
    """

    address: AccAddress = attr.ib()
    """Input / output address."""

    coins: Coins = attr.ib(converter=Coins)
    """Coins to be received."""

    def to_amino(self) -> dict:
        return {"address": self.address, "coins": self.coins.to_amino()}

    @classmethod
    def from_data(cls, data: dict):
        return cls(address=data["address"], coins=Coins.from_data(data["coins"]))

    def to_data(self) -> dict:
        return {"address": self.address, "coins": self.coins.to_data()}

    @classmethod
    def from_proto(cls, proto: Output_pb) -> MultiSendOutput:
        return cls(address=proto.address, coins=Coins.from_proto(proto.coins))

    def to_proto(self) -> Output_pb:
        return Output_pb(address=self.address, coins=self.coins.to_proto())


def convert_input_list(data: list) -> List[MultiSendInput]:
    if all(isinstance(x, MultiSendInput) for x in data):
        return data
    else:
        return [MultiSendInput(address=d["address"], coins=d["coins"]) for d in data]


def convert_output_list(data: list) -> List[MultiSendOutput]:
    if all(isinstance(x, MultiSendOutput) for x in data):
        return data
    else:
        return [MultiSendOutput(address=d["address"], coins=d["coins"]) for d in data]


@attr.s
class MsgMultiSend(Msg):
    """Allows batch-sending between multiple source and destination addresses.
    The total amount of coins in ``inputs`` must match ``outputs``. The transaction
    containing ``MsgMultiSend`` must contain signatures from all addresses used as inputs.

    Args:
        inputs (List[MultiSendInput]): senders and amounts
        outputs (List[MultiSendOutput]): recipients and amounts
    """

    type_amino = "cosmos-sdk/MsgMultiSend"
    """"""
    type_url = "/cosmos.bank.v1beta1.MsgMultiSend"
    """"""
    action = "multisend"
    """"""
    prototype = MsgMultiSend_pb
    """"""

    inputs: List[MultiSendInput] = attr.ib(converter=convert_input_list)
    outputs: List[MultiSendOutput] = attr.ib(converter=convert_output_list)

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "inputs": [mi.to_amino() for mi in self.inputs],
                "outputs": [mo.to_amino() for mo in self.outputs],
            },
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "inputs": [mi.to_data() for mi in self.inputs],
            "outputs": [mo.to_data() for mo in self.outputs],
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgMultiSend:
        return cls(
            inputs=[MultiSendInput.from_data(x) for x in data["inputs"]],
            outputs=[MultiSendOutput.from_data(x) for x in data["outputs"]],
        )

    @classmethod
    def from_proto(cls, proto: MsgMultiSend_pb) -> MsgMultiSend:
        return cls(
            inputs=[MultiSendInput.from_proto(x) for x in proto.inputs],
            outputs=[MultiSendOutput.from_proto(x) for x in proto.outputs],
        )

    def to_proto(self) -> MsgMultiSend_pb:
        return MsgMultiSend_pb(
            inputs=[i.to_proto() for i in self.inputs],
            outputs=[o.to_proto() for o in self.outputs],
        )

    @classmethod
    def unpack_any(cls, any_pb: Any_pb) -> MsgMultiSend:
        return cls.from_proto(MsgMultiSend_pb().parse(any_pb))



# core msgs
from terra_sdk.core.authz import (
    MsgExecAuthorized,
    MsgGrantAuthorization,
    MsgRevokeAuthorization,
)
from terra_sdk.core.bank import MsgMultiSend, MsgSend
from terra_sdk.core.crisis import MsgVerifyInvariant
from terra_sdk.core.distribution import (
    MsgFundCommunityPool,
    MsgSetWithdrawAddress,
    MsgWithdrawDelegatorReward,
    MsgWithdrawValidatorCommission,
)
from terra_sdk.core.feegrant import MsgGrantAllowance, MsgRevokeAllowance
from terra_sdk.core.gov.msgs import MsgDeposit, MsgSubmitProposal, MsgVote
from terra_sdk.core.ibc.msgs import (
    MsgAcknowledgement,
    MsgChannelCloseConfirm,
    MsgChannelCloseInit,
    MsgChannelOpenAck,
    MsgChannelOpenConfirm,
    MsgChannelOpenInit,
    MsgChannelOpenTry,
    MsgConnectionOpenAck,
    MsgConnectionOpenConfirm,
    MsgConnectionOpenInit,
    MsgConnectionOpenTry,
    MsgCreateClient,
    MsgRecvPacket,
    MsgSubmitMisbehaviour,
    MsgTimeout,
    MsgUpdateClient,
    MsgUpgradeClient,
)
from terra_sdk.core.ibc_transfer import MsgTransfer
from terra_sdk.core.slashing import MsgUnjail
from terra_sdk.core.staking import (
    MsgBeginRedelegate,
    MsgCreateValidator,
    MsgDelegate,
    MsgEditValidator,
    MsgUndelegate,
)
from terra_sdk.core.wasm import (
    MsgClearAdmin,
    MsgExecuteContract,
    MsgInstantiateContract,
    MsgMigrateContract,
    MsgStoreCode,
    MsgUpdateAdmin,
)

from .base import create_demux, create_demux_proto, create_demux_unpack_any

bank_msgs = [MsgSend, MsgMultiSend]
distribution_msgs = [
    MsgFundCommunityPool,
    MsgSetWithdrawAddress,
    MsgWithdrawDelegatorReward,
    MsgWithdrawValidatorCommission,
]
gov_msgs = [MsgDeposit, MsgSubmitProposal, MsgVote]
authz_msgs = [
    MsgExecAuthorized,
    MsgGrantAuthorization,
    MsgRevokeAuthorization,
]
slashing_msgs = [MsgUnjail]
staking_msgs = [
    MsgBeginRedelegate,
    MsgCreateValidator,
    MsgDelegate,
    MsgEditValidator,
    MsgUndelegate,
]
wasm_msgs = [
    MsgStoreCode,
    MsgInstantiateContract,
    MsgExecuteContract,
    MsgMigrateContract,
    MsgUpdateAdmin,
    MsgClearAdmin,
]
feegrant_msgs = [MsgGrantAllowance, MsgRevokeAllowance]

ibc_transfer_msgs = [MsgTransfer]
ibc_msgs = [
    MsgCreateClient,
    MsgUpdateClient,
    MsgUpgradeClient,
    MsgSubmitMisbehaviour,
    MsgConnectionOpenInit,
    MsgConnectionOpenTry,
    MsgConnectionOpenAck,
    MsgConnectionOpenConfirm,
    MsgChannelOpenInit,
    MsgChannelOpenTry,
    MsgChannelOpenAck,
    MsgChannelOpenConfirm,
    MsgChannelCloseInit,
    MsgChannelCloseConfirm,
    MsgRecvPacket,
    MsgTimeout,
    MsgAcknowledgement,
]
crisis_msgs = [MsgVerifyInvariant]

parse_msg = create_demux(
    [
        *authz_msgs,
        *bank_msgs,
        *distribution_msgs,
        *feegrant_msgs,
        *gov_msgs,
        *slashing_msgs,
        *staking_msgs,
        *wasm_msgs,
        *ibc_msgs,
        *ibc_transfer_msgs,
        *crisis_msgs,
    ]
)

parse_proto = create_demux_proto(
    [
        *authz_msgs,
        *bank_msgs,
        *distribution_msgs,
        *feegrant_msgs,
        *gov_msgs,
        *slashing_msgs,
        *staking_msgs,
        *wasm_msgs,
        *ibc_msgs,
        *ibc_transfer_msgs,
        *crisis_msgs,
    ]
)


parse_unpack_any = create_demux_unpack_any(
    [
        *authz_msgs,
        *bank_msgs,
        *distribution_msgs,
        *feegrant_msgs,
        *gov_msgs,
        *slashing_msgs,
        *staking_msgs,
        *wasm_msgs,
        *ibc_msgs,
        *ibc_transfer_msgs,
        *crisis_msgs,
    ]
)



"""ibc connection module message types."""

from __future__ import annotations

import attr
from betterproto.lib.google.protobuf import Any as Any_pb
from terra_proto.ibc.core.client.v1 import MsgCreateClient as MsgCreateClient_pb
from terra_proto.ibc.core.client.v1 import (
    MsgSubmitMisbehaviour as MsgSubmitMisbehaviour_pb,
)
from terra_proto.ibc.core.client.v1 import MsgUpdateClient as MsgUpdateClient_pb
from terra_proto.ibc.core.client.v1 import MsgUpgradeClient as MsgUpgradeClient_pb

from terra_sdk.core.msg import Msg

__all__ = [
    "MsgCreateClient",
    "MsgUpdateClient",
    "MsgUpgradeClient",
    "MsgSubmitMisbehaviour",
]


@attr.s
class MsgCreateClient(Msg):
    """
    MsgCreateClientResponse defines the Msg/CreateClient response type.
    """

    type_url = "/ibc.core.client.v1.MsgCreateClient"
    """"""
    prototype = MsgCreateClient_pb
    """"""

    client_state: dict = attr.ib()
    consensus_state: dict = attr.ib()
    signer: str = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> MsgCreateClient:
        return cls(
            client_state=data["client_state"],
            consensus_state=data["consensus_state"],
            signer=data["signer"],
        )

    def to_proto(self) -> MsgCreateClient_pb:
        return MsgCreateClient_pb(
            client_state=Any_pb().from_dict(self.client_state),
            consensus_state=Any_pb().from_dict(self.consensus_state),
            signer=self.signer,
        )

    @classmethod
    def from_proto(cls, proto: MsgCreateClient_pb) -> MsgCreateClient:
        return cls(
            client_state=proto.client_state.to_dict(),
            consensus_state=proto.consensus_state.to_dict(),
            signer=proto.signer,
        )


@attr.s
class MsgUpdateClient(Msg):
    """
    MsgUpdateClient defines a sdk.Msg to update an IBC client state using the given header.
    """

    type_url = "/ibc.core.client.v1.MsgUpdateClient"
    """"""
    prototype = MsgUpdateClient_pb
    """"""

    client_id: str = attr.ib()
    header: dict = attr.ib()
    signer: str = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> MsgUpdateClient:
        return cls(
            client_id=data["client_id"], header=data["header"], signer=data["signer"]
        )

    def to_proto(self) -> MsgUpdateClient_pb:
        return MsgUpdateClient_pb(
            client_id=self.client_id,
            header=Any_pb().from_dict(self.header),
            signer=self.signer,
        )

    @classmethod
    def from_proto(cls, proto: MsgUpdateClient_pb) -> MsgUpdateClient:
        return cls(
            client_id=proto.client_id,
            header=proto.header.to_dict(),
            signer=proto.signer,
        )


@attr.s
class MsgUpgradeClient(Msg):
    """
    MsgUpgradeClient defines an sdk.Msg to upgrade an IBC client to a new client state
    """

    type_url = "/ibc.core.client.v1.MsgUpgradeClient"
    """"""
    prototype = MsgUpgradeClient_pb
    """"""

    client_id: str = attr.ib()
    client_state: dict = attr.ib()
    consensus_state: dict = attr.ib()
    proof_upgrade_client: bytes = attr.ib()
    proof_upgrade_consensus_state: bytes = attr.ib()
    signer: str = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> MsgUpgradeClient:
        return cls(
            client_id=data["client_id"],
            client_state=data["client_state"],
            consensus_state=data["consensus_state"],
            proof_upgrade_client=data["proof_upgrade_client"],
            proof_upgrade_consensus_state=data["proof_upgrade_consensus_state"],
            signer=data["signer"],
        )

    def to_proto(self) -> MsgUpgradeClient_pb:
        return MsgUpgradeClient_pb(
            client_id=self.client_id,
            client_state=Any_pb().from_dict(self.client_state),
            consensus_state=Any_pb().from_dict(self.consensus_state),
            proof_upgrade_client=self.proof_upgrade_client,
            proof_upgrade_consensus_state=self.proof_upgrade_consensus_state,
            signer=self.signer,
        )

    @classmethod
    def from_proto(cls, proto: MsgUpgradeClient_pb) -> MsgUpgradeClient:
        return cls(
            client_id=proto.client_id,
            client_state=proto.client_state.to_dict(),
            consensus_state=proto.consensus_state.to_dict(),
            proof_upgrade_client=proto.proof_upgrade_client,
            proof_upgrade_consensus_state=proto.proof_upgrade_consensus_state,
            signer=proto.signer,
        )


@attr.s
class MsgSubmitMisbehaviour(Msg):
    """
    MsgSubmitMisbehaviour defines a sdk.Msg type that submits Evidence for light client misbehaviour.
    """

    type_url = "/ibc.core.client.v1.MsgSubmitMisbehaviour"
    """"""
    prototype = MsgSubmitMisbehaviour_pb
    """"""

    client_id: str = attr.ib()
    misbehaviour: any = attr.ib()
    signer: str = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> MsgSubmitMisbehaviour:
        return cls(
            client_id=data["client_id"],
            misbehaviour=data["misbehaviour"],
            signer=data["signer"],
        )

    def to_proto(self) -> MsgSubmitMisbehaviour_pb:
        return MsgSubmitMisbehaviour_pb(
            client_id=self.client_id,
            misbehaviour=Any_pb().from_dict(self.misbehaviour),
            signer=self.signer,
        )

    @classmethod
    def from_proto(cls, proto: MsgSubmitMisbehaviour_pb) -> MsgSubmitMisbehaviour:
        return cls(
            client_id=proto.client_id,
            misbehaviour=Any_pb().from_dict(proto.misbehaviour),
            signer=proto.signer,
        )



import base64
import copy
from typing import List, Optional

import attr
from multidict import CIMultiDict

from terra_sdk.core import AccAddress, Coins, Dec, Numeric, PublicKey
from terra_sdk.core.broadcast import (
    AsyncTxBroadcastResult,
    BlockTxBroadcastResult,
    SyncTxBroadcastResult,
)
from terra_sdk.core.fee import Fee
from terra_sdk.core.msg import Msg
from terra_sdk.core.tx import AuthInfo, SignerData, SignMode, Tx, TxBody, TxInfo
from terra_sdk.util.hash import hash_amino
from terra_sdk.util.json import JSONSerializable

from ..params import APIParams
from ._base import BaseAsyncAPI, sync_bind

__all__ = [
    "AsyncTxAPI",
    "TxAPI",
    "BroadcastOptions",
    "CreateTxOptions",
    "SignerOptions",
]


@attr.s
class SignerOptions:
    """SignerOptions specifies infomations about signers
    Args:
        address (AccAddress): address of the signer
        sequence (int, optional): nonce of the messages from the signer
        public_key (PublicKey, optional): signer's PublicKey
    """

    address: AccAddress = attr.ib()
    sequence: Optional[int] = attr.ib(default=None)
    public_key: Optional[PublicKey] = attr.ib(default=None)


@attr.s
class CreateTxOptions:
    """

    Args:
        msgs (List[Msg]): list of messages to include
        fee (Optional[Fee], optional): transaction fee. If ``None``, will be estimated.
            See more on `fee estimation`_.
        memo (str, optional): optional short string to include with transaction.
        gas (str, optional): gas limit to set per-transaction; set to "auto" to calculate sufficient gas automatically
        gas_prices (Coins.Input, optional): gas prices for fee estimation.
        gas_adjustment (Numeric.Input, optional): gas adjustment for fee estimation.
        fee_denoms (List[str], optional): list of denoms to use for fee after estimation.
        account_number (int, optional): account number (overrides blockchain query if
            provided)
        sequence (int, optional): sequence (overrides blockchain qu ery if provided)
        timeout_height (int, optional):  specifies a block timeout height to prevent the tx from being committed past a certain height.
        sign_mode: (SignMode, optional): SignMode.SIGN_MODE_DIRECT by default. multisig needs SignMode.SIGN_MODE_LEGACY_AMINO_JSON.
    """

    msgs: List[Msg] = attr.ib()
    fee: Optional[Fee] = attr.ib(default=None)
    memo: Optional[str] = attr.ib(default=None)
    gas: Optional[str] = attr.ib(default=None)
    gas_prices: Optional[Coins.Input] = attr.ib(default=None)
    gas_adjustment: Optional[Numeric.Output] = attr.ib(
        default=0, converter=Numeric.parse
    )
    fee_denoms: Optional[List[str]] = attr.ib(default=None)
    account_number: Optional[int] = attr.ib(default=None)
    sequence: Optional[int] = attr.ib(default=None)
    timeout_height: Optional[int] = attr.ib(default=None)
    sign_mode: Optional[SignMode] = attr.ib(default=None)


@attr.s
class BroadcastOptions:
    sequences: Optional[List[int]] = attr.ib()
    fee_granter: Optional[AccAddress] = attr.ib(default=None)


""" deprecated
@attr.s
class TxSearchOption:
    key: str = attr.ib()
    value: Union[str, int] = attr.ib()
"""


@attr.s
class GasInfo:
    gas_wanted: int = attr.ib(converter=int)
    gas_used: int = attr.ib(converter=int)


@attr.s
class EventAttribute:
    key: str = attr.ib()
    value: str = attr.ib()


@attr.s
class Event:
    type: str = attr.ib()
    attributes: List[EventAttribute] = attr.ib(converter=list)


@attr.s
class SimulateResult:
    data: str = attr.ib()
    log: str = attr.ib()
    events: List[Event] = attr.ib(converter=list)


@attr.s
class SimulateResponse(JSONSerializable):
    gas_info: GasInfo = attr.ib()
    result: SimulateResult = attr.ib()

    @classmethod
    def from_data(cls, data: dict):
        return cls(gas_info=data["gas_info"], result=data["result"])


class AsyncTxAPI(BaseAsyncAPI):
    async def tx_info(self, tx_hash: str) -> TxInfo:
        """Fetches information for an included transaction given a tx hash.

        Args:
            tx_hash (str): hash of transaction to lookup

        Returns:
            TxInfo: transaction info
        """
        res = await self._c._get(f"/cosmos/tx/v1beta1/txs/{tx_hash}")
        return TxInfo.from_data(res["tx_response"])

    async def create(
        self, signers: List[SignerOptions], options: CreateTxOptions
    ) -> Tx:
        """Create a new unsigned transaction, with helpful utilities such as lookup of
        chain ID, account number, sequence and fee estimation.

        Args:
            signers (List[SignerOptions]): options about signers
            options (CreateTxOptions): options about creating a tx

        Returns:
            Tx: unsigned tx
        """

        opt = copy.deepcopy(options)

        signer_data: List[SignerData] = []
        for signer in signers:
            seq = signer.sequence
            pubkey = signer.public_key

            if seq is None or pubkey is None:
                acc = await BaseAsyncAPI._try_await(
                    self._c.auth.account_info(signer.address)
                )
                if seq is None:
                    seq = acc.get_sequence()
                if pubkey is None:
                    pubkey = acc.get_public_key()
            signer_data.append(SignerData(seq, pubkey))

        # create the fake fee
        if opt.fee is None:
            opt.fee = await BaseAsyncAPI._try_await(self.estimate_fee(signer_data, opt))

        return Tx(
            TxBody(opt.msgs, opt.memo or "", opt.timeout_height or 0),
            AuthInfo([], opt.fee),
            [],
        )

    async def estimate_fee(
        self, signers: List[SignerOptions], options: CreateTxOptions
    ) -> Fee:
        """Estimates the proper fee to apply by simulating it within the node.

        Args:
            signers ([SignerOptions]): signers
            options (CreateTxOptions): transaction info to estimate fee

        Returns:
            Fee: estimated fee
        """

        gas_prices = options.gas_prices or self._c.gas_prices
        gas_adjustment = options.gas_adjustment or self._c.gas_adjustment

        gas_prices_coins = None
        if gas_prices:
            gas_prices_coins = Coins(gas_prices)
            if options.fee_denoms:
                _fee_denoms = options.fee_denoms if options.fee_denoms else ["uluna"]
                gas_prices_coins = gas_prices_coins.filter(
                    lambda c: c.denom in _fee_denoms
                )
        tx_body = TxBody(messages=options.msgs, memo=options.memo or "")
        emptyCoins = Coins()
        emptyFee = Fee(0, emptyCoins)
        auth_info = AuthInfo([], emptyFee)

        tx = Tx(tx_body, auth_info, [])
        tx.append_empty_signatures(signers)

        gas = options.gas
        if gas is None or gas == "auto" or int(gas) == 0:
            opt = copy.deepcopy(options)
            opt.gas_adjustment = gas_adjustment
            gas = str(await super()._try_await(self.estimate_gas(tx, opt)))

        fee_amount = (
            gas_prices_coins.mul(gas).to_int_ceil_coins()
            if gas_prices_coins
            else Coins.from_str("0uluna")
        )

        return Fee(Numeric.parse(gas), fee_amount, "", "")

    async def estimate_gas(self, tx: Tx, options: Optional[CreateTxOptions]) -> int:
        gas_adjustment = options.gas_adjustment if options else self._c.gas_adjustment

        res = await self._c._post(
            "/cosmos/tx/v1beta1/simulate",
            {"tx_bytes": await super()._try_await(self.encode(tx))},
        )
        simulated = SimulateResponse.from_data(res)

        return int(Dec(gas_adjustment).mul(simulated.gas_info["gas_used"]))

    async def encode(self, tx: Tx) -> str:
        """Encode a Tx to base64 encoded proto string"""
        return base64.b64encode(bytes(tx.to_proto())).decode()

    async def decode(self, tx: str) -> Tx:
        """Decode base64 encoded proto string to a Tx"""
        return Tx.from_bytes(base64.b64decode(tx))

    async def hash(self, tx: Tx) -> str:
        """Compute hash for a transaction.

        Args:
            tx (Tx): transaction to hash

        Returns:
            str: transaction hash
        """
        amino = await super()._try_await(self.encode(tx))
        return hash_amino(amino)

    async def _broadcast(
        self, tx: Tx, mode: str, options: BroadcastOptions = None
    ) -> dict:
        data = {"tx_bytes": await super()._try_await(self.encode(tx)), "mode": mode}
        return await self._c._post("/cosmos/tx/v1beta1/txs", data)  # , raw=True)

    async def broadcast_sync(
        self, tx: Tx, options: BroadcastOptions = None
    ) -> SyncTxBroadcastResult:
        """Broadcasts a transaction using the ``sync`` broadcast mode.

        Args:
            tx (Tx): transaction to broadcast
            options (BroadcastOptions): broacast options, optional

        Returns:
            SyncTxBroadcastResult: result
        """
        res = await self._broadcast(tx, "BROADCAST_MODE_SYNC", options)
        res = res.get("tx_response")
        return SyncTxBroadcastResult(
            txhash=res.get("txhash"),
            raw_log=res.get("raw_log"),
            code=res.get("code"),
            codespace=res.get("codespace"),
        )

    async def broadcast_async(
        self, tx: Tx, options: BroadcastOptions = None
    ) -> AsyncTxBroadcastResult:
        """Broadcasts a transaction using the ``async`` broadcast mode.

        Args:
            tx (Tx): transaction to broadcast
            options (BroadcastOptions): broacast options, optional

        Returns:
            AsyncTxBroadcastResult: result
        """
        res = await self._broadcast(tx, "BROADCAST_MODE_ASYNC", options)
        res = res.get("tx_response")
        return AsyncTxBroadcastResult(
            txhash=res.get("txhash"),
        )

    async def broadcast(
        self, tx: Tx, options: BroadcastOptions = None
    ) -> BlockTxBroadcastResult:
        """Broadcasts a transaction using the ``block`` broadcast mode.

        Args:
            tx (Tx): transaction to broadcast
            options (BroadcastOptions): broacast options, optional

        Returns:
            BlockTxBroadcastResult: result
        """
        res = await self._broadcast(tx, "BROADCAST_MODE_BLOCK", options)
        res = res["tx_response"]
        return BlockTxBroadcastResult(
            height=res.get("height") or 0,
            txhash=res.get("txhash"),
            raw_log=res.get("raw_log"),
            gas_wanted=res.get("gas_wanted") or 0,
            gas_used=res.get("gas_used") or 0,
            logs=res.get("logs"),
            code=res.get("code"),
            codespace=res.get("codespace"),
        )

    async def search(
        self, events: List[list], params: Optional[APIParams] = None
    ) -> dict:
        """Searches for transactions given criteria.

        Args:
            events (dict): dictionary containing options
            params (APIParams): optional parameters

        Returns:
            dict: transaction search results
        """

        actual_params = CIMultiDict()

        for event in events:
            if event[0] == "tx.height":
                actual_params.add("events", f"{event[0]}={event[1]}")
            else:
                actual_params.add("events", f"{event[0]}='{event[1]}'")
        if params:
            for p in params:
                actual_params.add(p, params[p])

        res = await self._c._get("/cosmos/tx/v1beta1/txs", actual_params)
        return {
            "txs": [TxInfo.from_data(tx) for tx in res.get("tx_responses")],
            "pagination": res.get("pagination"),
        }

    async def tx_infos_by_height(self, height: Optional[int] = None) -> List[TxInfo]:
        """Fetches information for an included transaction given block height or latest

        Args:
            height (int, optional): height to lookup. latest if height is None.

        Returns:
            List[TxInfo]: transaction info
        """
        if height is None:
            x = "latest"
        else:
            x = height

        res = await self._c._get(f"/cosmos/base/tendermint/v1beta1/blocks/{x}")

        txs = res.get("block").get("data").get("txs")
        hashes = [hash_amino(tx) for tx in txs]
        return [
            await BaseAsyncAPI._try_await(self.tx_info(tx_hash)) for tx_hash in hashes
        ]


class TxAPI(AsyncTxAPI):
    @sync_bind(AsyncTxAPI.tx_info)
    def tx_info(self, tx_hash: str) -> TxInfo:
        pass

    tx_info.__doc__ = AsyncTxAPI.tx_info.__doc__

    @sync_bind(AsyncTxAPI.create)
    def create(self, signers: List[SignerOptions], options: CreateTxOptions) -> Tx:
        pass

    create.__doc__ = AsyncTxAPI.create.__doc__

    @sync_bind(AsyncTxAPI.estimate_fee)
    def estimate_fee(
        self, signers: List[SignerOptions], options: CreateTxOptions
    ) -> Fee:
        pass

    estimate_fee.__doc__ = AsyncTxAPI.estimate_fee.__doc__

    @sync_bind(AsyncTxAPI.estimate_gas)
    def estimate_gas(
        self, tx: Tx, options: Optional[CreateTxOptions]
    ) -> SimulateResponse:
        pass

    estimate_gas.__doc__ = AsyncTxAPI.estimate_gas.__doc__

    @sync_bind(AsyncTxAPI.encode)
    def encode(self, tx: Tx) -> str:
        pass

    encode.__doc__ = AsyncTxAPI.encode.__doc__

    @sync_bind(AsyncTxAPI.decode)
    def decode(self, tx: str) -> Tx:
        pass

    decode.__doc__ = AsyncTxAPI.decode.__doc__

    @sync_bind(AsyncTxAPI.hash)
    def hash(self, tx: Tx) -> str:
        pass

    hash.__doc__ = AsyncTxAPI.hash.__doc__

    @sync_bind(AsyncTxAPI.broadcast_sync)
    def broadcast_sync(
        self, tx: Tx, options: BroadcastOptions = None
    ) -> SyncTxBroadcastResult:
        pass

    broadcast_sync.__doc__ = AsyncTxAPI.broadcast_sync.__doc__

    @sync_bind(AsyncTxAPI.broadcast_async)
    def broadcast_async(
        self, tx: Tx, options: BroadcastOptions = None
    ) -> AsyncTxBroadcastResult:
        pass

    broadcast_async.__doc__ = AsyncTxAPI.broadcast_async.__doc__

    @sync_bind(AsyncTxAPI.broadcast)
    def broadcast(
        self, tx: Tx, options: BroadcastOptions = None
    ) -> BlockTxBroadcastResult:
        pass

    broadcast.__doc__ = AsyncTxAPI.broadcast.__doc__

    @sync_bind(AsyncTxAPI.search)
    def search(self, events: List[list], params: Optional[APIParams] = None) -> dict:
        pass

    search.__doc__ = AsyncTxAPI.search.__doc__

    @sync_bind(AsyncTxAPI.tx_infos_by_height)
    def tx_infos_by_height(self, height: Optional[int] = None) -> List[TxInfo]:
        pass

    tx_infos_by_height.__doc__ = AsyncTxAPI.tx_infos_by_height.__doc__



from terra_proto.cosmos.gov.v1beta1 import ProposalStatus

from .data import Content, Proposal, VoteOption, WeightedVoteOption
from .msgs import MsgDeposit, MsgSubmitProposal, MsgVote
from .proposals import TextProposal

__all__ = [
    "Content",
    "MsgDeposit",
    "MsgSubmitProposal",
    "MsgVote",
    "Proposal",
    "TextProposal",
    "ProposalStatus",
    "VoteOption",
    "WeightedVoteOption",
]



from .data import (
    Commission,
    CommissionRates,
    Delegation,
    Description,
    Redelegation,
    RedelegationEntry,
    UnbondingDelegation,
    UnbondingDelegationEntry,
    Validator,
)
from .msgs import (
    MsgBeginRedelegate,
    MsgCreateValidator,
    MsgDelegate,
    MsgEditValidator,
    MsgUndelegate,
)

__all__ = [
    "Commission",
    "CommissionRates",
    "Delegation",
    "Description",
    "MsgBeginRedelegate",
    "MsgCreateValidator",
    "MsgDelegate",
    "MsgEditValidator",
    "MsgUndelegate",
    "Redelegation",
    "RedelegationEntry",
    "UnbondingDelegation",
    "UnbondingDelegationEntry",
    "Validator",
]



from typing import List, Optional, Union

from dateutil import parser

from terra_sdk.core import Dec, Numeric, ValConsPubKey

from ._base import BaseAsyncAPI, sync_bind

__all__ = ["AsyncSlashingAPI", "SlashingAPI"]

from ..params import APIParams


class AsyncSlashingAPI(BaseAsyncAPI):
    async def signing_info(
        self, val_cons_pub_key: ValConsPubKey
    ) -> Union[List[dict], dict]:
        """Fetches signing info for a validator consensus public key.

        Args:
            val_cons_pub_key (ValConsPubKey): validator consensus public key.

        Returns:
            Union[List[dict], dict]: signing info
        """
        res = await self._c._get(
            f"/cosmos/slashing/v1beta1/signing_infos/{val_cons_pub_key}"
        )
        info = res["val_signing_info"]
        return {
            "address": info["address"],
            "start_height": Numeric.parse(info["start_height"]),
            "index_offset": Numeric.parse(info["index_offset"]),
            "jailed_until": parser.parse(info["jailed_until"]),
            "tombstoned": bool(info["tombstoned"]),
            "missed_blocks_counter": Numeric.parse(info["missed_blocks_counter"]),
        }

    async def signing_infos(
        self, params: Optional[APIParams] = None
    ) -> (Union[List[dict], dict], dict):
        """Fetches all signing info.

        Args:
            params (APIParams, optional): additional params for the API like pagination

        Returns:
            Union[List[dict], dict]: signing infos
            dict: pagination info
        """
        res = await self._c._get("/cosmos/slashing/v1beta1/signing_infos", params)
        infos = res["info"]
        return [
            {
                "address": info["address"],
                "start_height": Numeric.parse(info["start_height"]),
                "index_offset": Numeric.parse(info["index_offset"]),
                "jailed_until": info["jailed_until"],  # TODO: convert to datetime
                "tombstoned": bool(info["tombstoned"]),
                "missed_blocks_counter": Numeric.parse(info["missed_blocks_counter"]),
            }
            for info in infos
        ], res.get("pagination")

    async def parameters(self) -> dict:
        """Fetches Slashing module parameters.

        Returns:
            dict: Slashing module parameters
        """
        res = await self._c._get("/cosmos/slashing/v1beta1/params")
        params = res.get("params")
        return {
            "signed_blocks_window": Numeric.parse(params["signed_blocks_window"]),
            "min_signed_per_window": Dec(params["min_signed_per_window"]),
            "downtime_jail_duration": params["downtime_jail_duration"],
            "slash_fraction_double_sign": Dec(params["slash_fraction_double_sign"]),
            "slash_fraction_downtime": Dec(params["slash_fraction_downtime"]),
        }


class SlashingAPI(AsyncSlashingAPI):
    @sync_bind(AsyncSlashingAPI.signing_info)
    def signing_info(self, val_cons_pub_key: ValConsPubKey) -> List[dict]:
        pass

    signing_info.__doc__ = AsyncSlashingAPI.signing_info.__doc__

    @sync_bind(AsyncSlashingAPI.signing_infos)
    def signing_infos(
        self, params: Optional[APIParams]
    ) -> (Union[List[dict], dict], dict):
        pass

    signing_infos.__doc__ = AsyncSlashingAPI.signing_infos.__doc__

    @sync_bind(AsyncSlashingAPI.parameters)
    def parameters(self) -> dict:
        pass

    parameters.__doc__ = AsyncSlashingAPI.parameters.__doc__



from terra_proto.cosmos.authz.v1beta1 import (
    GenericAuthorization as GenericAuthorization_pb,
)
from terra_proto.cosmos.bank.v1beta1 import SendAuthorization as SendAuthorization_pb
from terra_proto.cosmos.staking.v1beta1 import (
    StakeAuthorization as StakeAuthorization_pb,
)

from terra_sdk.core.authz import (
    GenericAuthorization,
    SendAuthorization,
    StakeAuthorization,
)

from .base import (
    create_demux,
    create_demux_amino,
    create_demux_proto,
    create_demux_unpack_any,
)

parse_authorization = create_demux(
    [GenericAuthorization, SendAuthorization, StakeAuthorization]
)  # data

parse_authorization_amino = create_demux_amino(
    [GenericAuthorization, SendAuthorization]
)  # no amino for StakeAuthorization

parse_authorization_unpack_any = create_demux_unpack_any(
    [GenericAuthorization, SendAuthorization, StakeAuthorization]
)

parse_authorization_proto = create_demux_proto(
    [
        GenericAuthorization,
        SendAuthorization,
        StakeAuthorization,
    ]
)



"""Crisis module message types."""

from __future__ import annotations

from betterproto.lib.google.protobuf import Any as Any_pb
from terra_proto.cosmos.crisis.v1beta1 import (
    MsgVerifyInvariant as MsgVerifyInvariant_pb,
)

from terra_sdk.core import AccAddress
from terra_sdk.core.msg import Msg

__all__ = ["MsgVerifyInvariant"]

import attr


@attr.s
class MsgVerifyInvariant(Msg):
    """MsgVerifyInvariant represents a message to verify a particular invariance.

    Args:
        sender: address of the sender
        invariant_module_name: module name to verify invariant
        invariant_route: route to veriryf
    """

    type_amino = "cosmos-sdk/MsgVerifyInvariant"
    """"""
    type_url = "/cosmos.crisis.v1beta1.MsgVerifyInvariant"
    """"""
    prototype = MsgVerifyInvariant_pb
    """"""

    sender: AccAddress = attr.ib()
    invariant_module_name: str = attr.ib()
    invariant_route: str = attr.ib()

    def to_amino(self) -> dict:
        raise Exception("MsgVerifyInvarant is not allowed to send")

    @classmethod
    def from_data(cls, data: dict) -> MsgVerifyInvariant:
        return cls(
            sender=data["sender"],
            invariant_module_name=data["invariant_module_name"],
            invariant_route=data["invariant_route"],
        )

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "sender": self.sender,
            "invariant_module_name": self.invariant_module_name,
            "invariant_route": self.invariant_route,
        }

    @classmethod
    def from_proto(cls, proto: MsgVerifyInvariant_pb) -> MsgVerifyInvariant:
        return cls(
            sender=proto.sender,
            invariant_module_name=proto.invariant_module_name,
            invariant_route=proto.invariant_route,
        )

    def to_proto(self) -> MsgVerifyInvariant_pb:
        raise Exception("MsgVerifyInvarant is not allowed to send")

    @classmethod
    def unpack_any(cls, any_pb: Any_pb) -> MsgVerifyInvariant:
        return MsgVerifyInvariant.from_proto(any_pb)



"""ibc connection module message types."""

from __future__ import annotations

import attr
from terra_proto.ibc.core.channel.v1 import MsgAcknowledgement as MsgAcknowledgement_pb
from terra_proto.ibc.core.channel.v1 import (
    MsgChannelCloseConfirm as MsgChannelCloseConfirm_pb,
)
from terra_proto.ibc.core.channel.v1 import (
    MsgChannelCloseInit as MsgChannelCloseInit_pb,
)
from terra_proto.ibc.core.channel.v1 import MsgChannelOpenAck as MsgChannelOpenAck_pb
from terra_proto.ibc.core.channel.v1 import (
    MsgChannelOpenConfirm as MsgChannelOpenConfirm_pb,
)
from terra_proto.ibc.core.channel.v1 import MsgChannelOpenInit as MsgChannelOpenInit_pb
from terra_proto.ibc.core.channel.v1 import MsgChannelOpenTry as MsgChannelOpenTry_pb
from terra_proto.ibc.core.channel.v1 import MsgRecvPacket as MsgRecvPacket_pb
from terra_proto.ibc.core.channel.v1 import MsgTimeout as MsgTimeout_pb

from terra_sdk.core.ibc.data.channel import Channel, Packet
from terra_sdk.core.ibc.data.client import Height
from terra_sdk.core.msg import Msg

# TODO: support MsgTimeoutClose
__all__ = [
    "MsgChannelOpenInit",
    "MsgChannelOpenTry",
    "MsgChannelOpenAck",
    "MsgChannelOpenConfirm",
    "MsgChannelCloseInit",
    "MsgChannelCloseConfirm",
    "MsgRecvPacket",
    "MsgTimeout",
    "MsgAcknowledgement",
]


@attr.s
class MsgChannelOpenInit(Msg):
    """
    MsgChannelOpenInit defines an sdk.Msg to initialize a channel handshake. It
    is called by a relayer on Chain A.
    """

    type_url = "/ibc.core.channel.v1.MsgChannelOpenInit"
    """"""
    prototype = MsgChannelOpenInit_pb
    """"""

    port_id: str = attr.ib()
    channel: Channel = attr.ib()
    signer: str = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> MsgChannelOpenInit:
        return cls(
            port_id=data["port_id"],
            channel=Channel.from_data(data["port_id"]),
            signer=data["signer"],
        )

    def to_proto(self) -> MsgChannelOpenInit_pb:
        return MsgChannelOpenInit_pb(
            port_id=self.port_id, channel=self.channel.to_proto(), signer=self.signer
        )

    @classmethod
    def from_proto(cls, proto: MsgChannelOpenInit_pb) -> MsgChannelOpenInit:
        return cls(
            port_id=proto.port_id,
            channel=Channel.from_proto(proto.channel),
            signer=proto.signer,
        )


@attr.s
class MsgChannelOpenTry(Msg):
    """
    MsgChannelOpenInit defines a msg sent by a Relayer to try to open a channel
    on Chain B.
    """

    type_url = "/ibc.core.channel.v1.MsgChannelOpenTry"
    """"""
    prototype = MsgChannelOpenTry_pb
    """"""

    port_id: str = attr.ib()
    previous_channel_id: str = attr.ib()
    channel: Channel = attr.ib()
    counterparty_version: str = attr.ib()
    proof_init: bytes = attr.ib()
    proof_height: Height = attr.ib()
    signer: str = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> MsgChannelOpenTry:
        return cls(
            port_id=data["port_id"],
            previous_channel_id=data["previous_channel_id"],
            channel=Channel.from_data(data["channel"]),
            counterparty_version=data["counterparty_version"],
            proof_init=data["proof_init"],
            proof_height=Height.from_data(data["proof_height"]),
            signer=data["signer"],
        )

    def to_proto(self) -> MsgChannelOpenTry_pb:
        return MsgChannelOpenTry_pb(
            port_id=self.port_id,
            previous_channel_id=self.previous_channel_id,
            channel=self.channel.to_proto(),
            counterparty_version=self.counterparty_version,
            proof_init=self.proof_init,
            proof_height=self.proof_height.to_proto(),
            signer=self.signer,
        )

    @classmethod
    def from_proto(cls, proto: MsgChannelOpenTry_pb) -> MsgChannelOpenTry:
        return cls(
            port_id=proto.port_id,
            previous_channel_id=proto.previous_channel_id,
            channel=Channel.from_proto(proto.channel),
            counterparty_version=proto.counterparty_version,
            proof_init=proto.proof_init,
            proof_height=Height.from_proto(proto.proof_height),
            signer=proto.signer,
        )


class MsgChannelOpenAck(Msg):
    """
    MsgChannelOpenAck defines a msg sent by a Relayer to Chain A to acknowledge
    the change of channel state to TRYOPEN on Chain B.
    """

    type_url = "/ibc.core.channel.v1.MsgChannelOpenAck"
    """"""
    prototype = MsgChannelOpenAck_pb
    """"""

    port_id: str = attr.ib()
    channel_id: str = attr.ib()
    counterparty_channel_id: str = attr.ib()
    counterparty_version: str = attr.ib()
    proof_try: bytes = attr.ib()
    proof_height: Height = attr.ib()
    signer: str = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> MsgChannelOpenAck:
        return cls(
            port_id=data["port_id"],
            channel_id=data["channel_id"],
            counterparty_channel_id=data["counterparty_channel_id"],
            counterparty_version=data["counterparty_version"],
            proof_try=data["proof_try"],
            proof_height=Height.from_data(data["proof_height"]),
            signer=data["signer"],
        )

    def to_proto(self) -> MsgChannelOpenAck_pb:
        return MsgChannelOpenAck_pb(
            port_id=self.port_id,
            channel_id=self.channel_id,
            counterparty_channel_id=self.counterparty_channel_id,
            counterparty_version=self.counterparty_version,
            proof_try=self.proof_try,
            proof_height=self.proof_height.to_proto(),
            signer=self.signer,
        )

    @classmethod
    def from_proto(cls, proto: MsgChannelOpenAck) -> MsgChannelOpenAck:
        return cls(
            port_id=proto.port_id,
            channel_id=proto.channel_id,
            counterparty_channel_id=proto.counterparty_channel_id,
            counterparty_version=proto.counterparty_version,
            proof_try=proto.proof_try,
            proof_height=Height.from_proto(proto.proof_height),
            signer=proto.signer,
        )


@attr.s
class MsgChannelOpenConfirm(Msg):
    """
    MsgChannelOpenConfirm defines a msg sent by a Relayer to Chain B to
    acknowledge the change of channel state to OPEN on Chain A.
    """

    type_url = "/ibc.core.channel.v1.MsgChannelOpenConfirm"
    """"""
    prototype = MsgChannelOpenConfirm_pb
    """"""

    port_id: str = attr.ib()
    channel_id: str = attr.ib()
    proof_ack: bytes = attr.ib()
    proof_height: Height = attr.ib()
    signer: str = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> MsgChannelOpenConfirm:
        return cls(
            port_id=data["port_id"],
            channel_id=data["channel_id"],
            proof_ack=data["proof_ack"],
            proof_height=Height.from_data(data["proof_height"]),
            signer=data["signer"],
        )

    def to_proto(self) -> MsgChannelOpenConfirm_pb:
        return MsgChannelOpenConfirm_pb(
            port_id=self.port_id,
            channel_id=self.channel_id,
            proof_ack=self.proof_ack,
            proof_height=self.proof_height.to_proto(),
            signer=self.signer,
        )

    @classmethod
    def from_proto(cls, proto: MsgChannelOpenConfirm_pb) -> MsgChannelOpenConfirm:
        return cls(
            port_id=proto.port_id,
            channel_id=proto.channel_id,
            proof_ack=proto.proof_ack,
            proof_height=Height.from_proto(proto.proof_height),
            signer=proto.signer,
        )


@attr.s
class MsgChannelCloseInit(Msg):
    """ """

    type_url = "/ibc.core.channel.v1.MsgChannelCloseInit"
    """"""
    prototype = MsgChannelCloseInit_pb
    """"""

    port_id: str = attr.ib()
    channel_id: str = attr.ib()
    signer: str = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> MsgChannelCloseInit:
        return cls(
            port_id=data["port_id"],
            channel_id=data["channel_id"],
            signer=data["signer"],
        )

    def to_proto(self) -> MsgChannelCloseInit_pb:
        return MsgChannelCloseInit_pb(
            port_id=self.port_id, channel_id=self.channel_id, signer=self.signer
        )

    @classmethod
    def from_proto(cls, proto: MsgChannelOpenInit_pb) -> MsgChannelCloseInit:
        return cls(
            port_id=proto.port_id, channel_id=proto.channel_id, signer=proto.signer
        )


@attr.s
class MsgChannelCloseConfirm(Msg):
    """
    MsgChannelCloseConfirm defines a msg sent by a Relayer to Chain B to
    acknowledge the change of channel state to CLOSED on Chain A.
    """

    type_url = "/ibc.core.channel.v1.MsgChannelCloseConfirm"
    """"""
    prototype = MsgChannelCloseConfirm_pb
    """"""

    port_id: str = attr.ib()
    channel_id: str = attr.ib()
    proof_init: bytes = attr.ib()
    proof_height: Height = attr.ib()
    signer: str = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> MsgChannelCloseConfirm:
        return cls(
            port_id=data["port_id"],
            channel_id=data["channel_id"],
            proof_init=data["proof_init"],
            proof_height=Height.from_data(data["proof_height"]),
            signer=data["signer"],
        )

    def to_proto(self) -> MsgChannelCloseConfirm_pb:
        return MsgChannelCloseConfirm_pb(
            port_id=self.port_id,
            channel_id=self.channel_id,
            proof_init=self.proof_init,
            proof_height=self.proof_height.to_proto(),
            signer=self.signer,
        )

    @classmethod
    def from_proto(cls, proto: MsgChannelCloseConfirm_pb) -> MsgChannelCloseConfirm:
        return cls(
            port_id=proto.port_id,
            channel_id=proto.channel_id,
            proof_init=proto.proof_init,
            proof_height=Height.from_proto(proto.proof_height),
            signer=proto.signer,
        )


@attr.s
class MsgRecvPacket(Msg):
    """
    MsgRecvPacket receives incoming IBC packet
    """

    type_url = "/ibc.core.channel.v1.MsgRecvPacket"
    """"""
    prototype = MsgRecvPacket_pb
    """"""

    packet: Packet = attr.ib()
    proof_commitment: bytes = attr.ib()
    proof_height: Height = attr.ib()
    signer: str = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> MsgRecvPacket:
        return cls(
            packet=Packet.from_data(data["packet"]),
            proof_commitment=data["proof_commitment"],
            proof_height=Height.from_data(data["proof_height"]),
            signer=data["signer"],
        )

    def to_proto(self) -> MsgRecvPacket_pb:
        return MsgRecvPacket_pb(
            packet=self.packet.to_proto(),
            proof_commitment=self.proof_commitment,
            proof_height=self.proof_height.to_proto(),
            signer=self.signer,
        )

    @classmethod
    def from_proto(cls, proto: MsgRecvPacket_pb) -> MsgRecvPacket:
        return cls(
            packet=Packet.from_proto(proto.packet),
            proof_commitment=proto.proof_commitment,
            proof_height=Height.from_proto(proto.proof_height),
            signer=proto.signer,
        )


@attr.s
class MsgTimeout(Msg):
    """
    MsgTimeout receives timed-out packet
    """

    type_url = "/ibc.core.channel.v1.MsgTimeout"
    """"""
    prototype = MsgTimeout_pb
    """"""

    packet: Packet = attr.ib()
    proof_unreceived: bytes = attr.ib()
    proof_height: Height = attr.ib()
    next_sequence_recv: int = attr.ib(converter=int)
    signer: str = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> MsgTimeout:
        return cls(
            packet=Packet.from_data(data["packet"]),
            proof_unreceived=data["proof_unreceived"],
            proof_height=Height.from_data(data["proof_height"]),
            next_sequence_recv=data["next_sequence_recv"],
            signer=data["signer"],
        )

    def to_proto(self) -> MsgTimeout_pb:
        return MsgTimeout_pb(
            packet=self.packet.to_proto(),
            proof_unreceived=self.proof_unreceived,
            proof_height=self.proof_height.to_proto(),
            next_sequence_recv=self.next_sequence_recv,
            signer=self.signer,
        )

    @classmethod
    def from_proto(cls, proto: MsgTimeout_pb) -> MsgTimeout:
        return cls(
            packet=Packet.from_proto(proto.packet),
            proof_unreceived=proto.proof_unreceived,
            proof_height=Height.from_proto(proto.proof_height),
            next_sequence_recv=proto.next_sequence_recv,
            signer=proto.signer,
        )


@attr.s
class MsgAcknowledgement(Msg):
    """
    MsgAcknowledgement receives incoming IBC acknowledgement
    """

    type_url = "/ibc.core.channel.v1.MsgAcknowledgement"
    """"""
    prototype = MsgAcknowledgement_pb
    """"""

    packet: Packet = attr.ib()
    acknowledgement: bytes = attr.ib()
    proof_acked: bytes = attr.ib()
    proof_height: Height = attr.ib()
    signer: str = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> MsgAcknowledgement:
        return cls(
            packet=Packet.from_data(data["packet"]),
            acknowledgement=data["acknowledgement"],
            proof_acked=data["proof_acked"],
            proof_height=Height.from_data(data["proof_height"]),
            signer=data["signer"],
        )

    def to_proto(self) -> MsgAcknowledgement_pb:
        return MsgAcknowledgement_pb(
            packet=self.packet.to_proto(),
            acknowledgement=self.acknowledgement,
            proof_acked=self.proof_acked,
            proof_height=self.proof_height.to_proto(),
            signer=self.signer,
        )

    @classmethod
    def from_proto(cls, proto: MsgAcknowledgement_pb) -> MsgAcknowledgement:
        return cls(
            packet=Packet.from_proto(proto.packet),
            acknowledgement=proto.acknowledgement,
            proof_acked=proto.proof_acked,
            proof_height=Height.from_proto(proto.proof_height),
            signer=proto.signer,
        )



"""ibc-trasfer module data objects."""
from __future__ import annotations

import attr
from terra_proto.ibc.applications.transfer.v1 import DenomTrace as DenomTrace_pb

from terra_sdk.util.json import JSONSerializable

__all__ = ["DenomTrace"]


@attr.s
class DenomTrace(JSONSerializable):
    path: str = attr.ib()
    base_denom: str = attr.ib()

    def to_amino(self) -> dict:
        return {"path": self.path, "base_denom": self.base_denom}

    @classmethod
    def from_data(cls, data: dict) -> DenomTrace:
        return cls(path=data["path"], base_denom=data["base_denom"])

    def to_proto(self) -> DenomTrace_pb:
        return DenomTrace_pb(path=self.path, base_denom=self.base_denom)



from .lcdclient import AsyncLCDClient, LCDClient
from .params import PaginationOptions
from .wallet import AsyncWallet, Wallet

__all__ = ["AsyncLCDClient", "LCDClient", "AsyncWallet", "Wallet", "PaginationOptions"]



from __future__ import annotations

from terra_sdk.key.key import Key, SignOptions

from .api.tx import CreateTxOptions, SignerOptions

__all__ = ["Wallet", "AsyncWallet"]

from ...core.tx import SignMode, Tx


class AsyncWallet:
    def __init__(self, lcd, key: Key):
        self.lcd = lcd
        self.key = key

    async def account_number(self) -> int:
        res = await self.lcd.auth.account_info(self.key.acc_address)
        return res.account_number

    async def sequence(self) -> int:
        res = await self.lcd.auth.account_info(self.key.acc_address)
        return res.sequence

    async def account_number_and_sequence(self) -> dict:
        res = await self.lcd.auth.account_info(self.key.acc_address)
        return {"account_number": res.account_number, "sequence": res.sequence}

    async def create_tx(self, options: CreateTxOptions) -> Tx:
        sigOpt = [
            SignerOptions(
                address=self.key.acc_address,
                sequence=options.sequence,
                public_key=self.key.public_key,
            )
        ]
        return await self.lcd.tx.create(sigOpt, options)

    async def create_and_sign_tx(self, options: CreateTxOptions) -> Tx:
        account_number = options.account_number
        sequence = options.sequence
        if account_number is None or sequence is None:
            res = await self.account_number_and_sequence()
            if account_number is None:
                account_number = res.get("account_number")
            if sequence is None:
                sequence = res.get("sequence")
        options.sequence = sequence
        options.account_number = account_number
        return self.key.sign_tx(
            tx=(await self.create_tx(options)),
            options=SignOptions(
                account_number=account_number,
                sequence=sequence,
                chain_id=self.lcd.chain_id,
                sign_mode=options.sign_mode
                if options.sign_mode
                else SignMode.SIGN_MODE_DIRECT,
            ),
        )


class Wallet:
    """Wraps around a :class:`Key` implementation and provides transaction building and
    signing functionality. It is recommended to create this object through
    :meth:`LCDClient.wallet()<terra_sdk.client.lcd.LCDClient.wallet>`."""

    def __init__(self, lcd, key: Key):
        self.lcd = lcd
        self.key = key

    def account_number(self) -> int:
        """Fetches account number for the account associated with the Key."""
        res = self.lcd.auth.account_info(self.key.acc_address)
        return res.account_number

    def sequence(self) -> int:
        """Fetches the sequence number for the account associated with the Key."""
        res = self.lcd.auth.account_info(self.key.acc_address)
        return res.sequence

    def account_number_and_sequence(self) -> dict:
        """Fetches both account and sequence number associated with the Key."""
        res = self.lcd.auth.account_info(self.key.acc_address)
        return {"account_number": res.account_number, "sequence": res.sequence}

    def create_tx(self, options: CreateTxOptions) -> Tx:
        """Builds an unsigned transaction object. The ``Wallet`` will first
        query the blockchain to fetch the latest ``account`` and ``sequence`` values for the
        account corresponding to its Key, unless the they are both provided. If no ``fee``
        parameter is set, automatic fee estimation will be used (see `fee_estimation`).

        Args:
            options (CreateTxOptions): Options to create a tx

        Returns:
            Tx: unsigned transaction
        """
        sigOpt = [
            SignerOptions(
                address=self.key.acc_address,
                sequence=options.sequence,
                public_key=self.key.public_key,
            )
        ]
        return self.lcd.tx.create(sigOpt, options)

    def create_and_sign_tx(self, options: CreateTxOptions) -> Tx:
        """Creates and signs a :class:`Tx` object in a single step. This is the recommended
        method for preparing transaction for immediate signing and broadcastring. The transaction
        is generated exactly as :meth:`create_tx`.

        Args:
            options (CreateTxOptions): Options to create a tx

        Returns:
            Tx: signed transaction
        """

        account_number = options.account_number
        sequence = options.sequence
        if account_number is None or sequence is None:
            res = self.account_number_and_sequence()
            if account_number is None:
                account_number = res.get("account_number")
            if sequence is None:
                sequence = res.get("sequence")
        options.sequence = sequence
        options.account_number = account_number
        return self.key.sign_tx(
            tx=self.create_tx(options),
            options=SignOptions(
                account_number=account_number,
                sequence=sequence,
                chain_id=self.lcd.chain_id,
                sign_mode=options.sign_mode
                if options.sign_mode
                else SignMode.SIGN_MODE_DIRECT,
            ),
        )



"""Gov module governance proposal types."""

from __future__ import annotations

import attr
from betterproto.lib.google.protobuf import Any as Any_pb
from terra_proto.cosmos.gov.v1beta1 import TextProposal as TextProposal_pb

from terra_sdk.util.json import JSONSerializable

__all__ = ["TextProposal"]


@attr.s
class TextProposal(JSONSerializable):
    """Generic proposal type with only title and description that does nothing if
    passed. Primarily used for assessing the community sentiment around the proposal.

    Args:
        title: proposal title
        description: proposal description
    """

    type_amino = "cosmos-sdk/TextProposal"
    """"""
    type_url = "/cosmos.gov.v1beta1.TextProposal"
    """"""
    prototype = TextProposal_pb
    """"""

    title: str = attr.ib()
    description: str = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {"title": self.title, "description": self.description},
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "title": self.title,
            "description": self.description,
        }

    @classmethod
    def from_data(cls, data: dict) -> TextProposal:
        return cls(title=data["title"], description=data["description"])

    def to_proto(self) -> TextProposal_pb:
        return TextProposal_pb(title=self.title, description=self.description)

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))

    @classmethod
    def from_proto(cls, proto: TextProposal_pb):
        return cls(title=proto.title, description=proto.description)



"""feegrant module data objects."""
from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from typing import List, Optional, Union

import attr
from attr import converters
from betterproto.lib.google.protobuf import Any as Any_pb
from dateutil import parser
from terra_proto.cosmos.feegrant.v1beta1 import (
    AllowedMsgAllowance as AllowedMsgAllowance_pb,
)
from terra_proto.cosmos.feegrant.v1beta1 import BasicAllowance as BasicAllowance_pb
from terra_proto.cosmos.feegrant.v1beta1 import (
    PeriodicAllowance as PeriodicAllowance_pb,
)

from terra_sdk.core import Coins
from terra_sdk.util.base import create_demux_proto, create_demux_unpack_any
from terra_sdk.util.converter import to_isoformat
from terra_sdk.util.json import JSONSerializable

__all__ = ["BasicAllowance", "PeriodicAllowance", "AllowedMsgAllowance", "Allowance"]


@attr.s
class BasicAllowance(JSONSerializable):
    """
    BasicAllowance implements Allowance with a one-time grant of tokens
    that optionally expires. The grantee can use up to SpendLimit to cover fees.
    """

    spend_limit: Optional[Coins] = attr.ib(converter=converters.optional(Coins))
    expiration: Optional[datetime] = attr.ib(converter=parser.parse)

    type_amino = "cosmos-sdk/BasicAllowance"
    """"""
    type_url = "/cosmos.feegrant.v1beta1.BasicAllowance"
    """"""
    prototype = BasicAllowance_pb
    """"""

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "spend_limit": self.spend_limit.to_amino()
                if self.spend_limit
                else None,
                "expiration": to_isoformat(self.expiration)
                if self.expiration
                else None,
            },
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "spend_limit": self.spend_limit.to_data() if self.spend_limit else None,
            "expiration": to_isoformat(self.expiration) if self.expiration else None,
        }

    @classmethod
    def from_amino(cls, amino: dict) -> BasicAllowance:
        data = amino.get("value")
        sl = data.get("spend_limit") or None
        exp = data.get("expiration") or None
        return cls(
            spend_limit=Coins.from_amino(sl) if sl else None,
            expiration=exp if exp else None,
        )

    @classmethod
    def from_data(cls, data: dict) -> BasicAllowance:
        sl = data.get("spend_limit")
        exp = data.get("expiration")
        return cls(
            spend_limit=Coins.from_data(sl) if sl else None,
            expiration=exp if exp else None,
        )

    def to_proto(self) -> BasicAllowance_pb:
        return BasicAllowance_pb(
            spend_limit=self.spend_limit.to_proto()
            if (self.spend_limit and len(self.spend_limit) > 0)
            else None,
            expiration=self.expiration,
        )

    @classmethod
    def from_proto(cls, proto: BasicAllowance_pb) -> BasicAllowance:
        sl = proto.spend_limit
        return cls(
            spend_limit=Coins.from_proto(sl) if sl else None,
            expiration=to_isoformat(proto.expiration),
        )


@attr.s
class PeriodicAllowance(JSONSerializable):
    """
    PeriodicAllowance extends Allowance to allow for both a maximum cap,
     as well as a limit per time period.
    """

    basic: BasicAllowance = attr.ib()
    period: int = attr.ib(converter=int)
    period_spend_limit: Coins = attr.ib(converter=Coins)
    period_can_spend: Coins = attr.ib(converter=Coins)
    period_reset: datetime = attr.ib(converter=parser.parse)

    type_amino = "cosmos-sdk/PeriodicAllowance"
    """"""
    type_url = "/cosmos.feegrant.v1beta1.PeriodicAllowance"
    """"""
    prototype = PeriodicAllowance_pb
    """"""

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "basic": self.basic.to_amino(),
                "period": str(self.period),
                "period_spend_limit": self.period_spend_limit.to_amino(),
                "period_can_spend": self.period_can_spend.to_amino(),
                "period_reset": to_isoformat(self.period_reset),
            },
        }

    @classmethod
    def from_amino(cls, amino: dict) -> PeriodicAllowance:
        data = amino.get("value")
        return cls(
            basic=BasicAllowance.from_amino(data.get("basic")),
            period=int(data.get("period")),
            period_spend_limit=Coins.from_amino(data.get("period_spend_limit")),
            period_can_spend=Coins.from_amino(data.get("period_can_spend")),
            period_reset=data.get("period_reset"),
        )

    @classmethod
    def from_data(cls, data: dict) -> PeriodicAllowance:
        return cls(
            basic=BasicAllowance.from_data(data["basic"]),
            period=data["period"],
            period_spend_limit=Coins.from_data(data["period_spend_limit"]),
            period_can_spend=Coins.from_data(data["period_can_spend"]),
            period_reset=data["period_reset"],
        )

    def to_proto(self) -> PeriodicAllowance_pb:
        return PeriodicAllowance_pb(
            basic=self.basic.to_proto(),
            period=timedelta(seconds=self.period),
            period_spend_limit=self.period_spend_limit.to_proto(),
            period_can_spend=self.period_can_spend.to_proto(),
            period_reset=self.period_reset,
        )

    @classmethod
    def from_proto(cls, proto: PeriodicAllowance_pb) -> PeriodicAllowance:
        return cls(
            basic=BasicAllowance.from_proto(proto.basic),
            period=proto.period.seconds(),
            period_spend_limit=proto.period_spend_limit,
            period_can_spend=proto.period_can_spend,
            period_reset=to_isoformat(proto.period_reset),
        )


@attr.s
class AllowedMsgAllowance(JSONSerializable):
    """
    AllowedMsgAllowance creates allowance only for specified message types.
    """

    allowance: Allowance = attr.ib()
    allowed_messages: List[str] = attr.ib(converter=list)

    type_amino = "cosmos-sdk/AllowedMsgAllowance"
    """"""
    type_url = "/cosmos.feegrant.v1beta1.AllowedMsgAllowance"
    """"""

    prototype = AllowedMsgAllowance_pb

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "allowance": self.allowance.to_amino(),
                "allowed_messages": self.allowed_messages,
            },
        }

    @classmethod
    def from_amino(cls, data: dict) -> AllowedMsgAllowance:
        allowance = data["allowance"]
        return cls(
            allowance=Allowance.from_amino(allowance),
            allowed_messages=data["allowed_messages"],
        )

    @classmethod
    def from_data(cls, data: dict) -> AllowedMsgAllowance:
        allowance = data["allowance"]
        return cls(
            allowance=Allowance.from_data(allowance),
            allowed_messages=data["allowed_messages"],
        )

    def to_proto(self) -> AllowedMsgAllowance_pb:
        return AllowedMsgAllowance_pb(
            allowance=Allowance.pack_any(self.allowance),
            allowed_messages=self.allowed_messages,
        )

    @classmethod
    def from_proto(cls, proto: AllowedMsgAllowance_pb) -> AllowedMsgAllowance:
        return cls(
            allowance=Allowance.unpack_any(proto.allowance),
            allowed_messages=proto.allowed_messages,
        )


class Allowance(JSONSerializable, ABC):  # (BasicAllowance, PeriodicAllowance):
    @property
    @abstractmethod
    def type_url(self):
        pass

    @property
    @abstractmethod
    def type_amino(self):
        pass

    @abstractmethod
    def to_amino(self) -> dict:
        pass

    @abstractmethod
    def to_data(self) -> dict:
        pass

    @abstractmethod
    def to_proto(self) -> dict:
        pass

    @classmethod
    def from_data(cls, data: dict):
        if data.get("@type") == BasicAllowance.type_url:
            return BasicAllowance.from_data(data)
        elif data.get("@type") == PeriodicAllowance.type_url:
            return PeriodicAllowance.from_data(data)
        elif data.get("@type") == AllowedMsgAllowance.type_url:
            return AllowedMsgAllowance.from_data(data)

    @classmethod
    def from_amino(cls, data: dict):
        if data.get("type") == BasicAllowance.type_amino:
            return BasicAllowance.from_amino(data)
        elif data.get("type") == PeriodicAllowance.type_amino:
            return PeriodicAllowance.from_amino(data)
        elif data.get("type") == AllowedMsgAllowance.type_amino:
            return AllowedMsgAllowance.from_amino(data)

    @classmethod
    def pack_any(
        cls, proto: Union[BasicAllowance, PeriodicAllowance, AllowedMsgAllowance]
    ) -> Any_pb:
        return Any_pb(type_url=proto.type_url, value=bytes(proto.to_proto()))

    @classmethod
    def unpack_any(cls, proto: Any_pb):
        return parse_allowance_unpack_any(proto)


parse_allowance_unpack_any = create_demux_unpack_any(
    [BasicAllowance, PeriodicAllowance, AllowedMsgAllowance]
)



"""ibc-transfer module message types."""

from __future__ import annotations

import attr
from terra_proto.ibc.applications.transfer.v1 import MsgTransfer as MsgTransfer_pb

from terra_sdk.core import AccAddress, Coin
from terra_sdk.core.ibc.data import Height
from terra_sdk.core.msg import Msg

__all__ = ["MsgTransfer"]


@attr.s
class MsgTransfer(Msg):
    """
    MsgTransfer defines a msg to transfer fungible tokens (i.e Coins) between ICS20 enabled chains.

    Args:
        source_port (str): the port on which the packet will be sent
        source_channel (str): the channel by which the packet will be sent
        token (Coin): the tokens to be transferred
        sender (AccAddress): the sender address
        receiver (str): the recipient address on the destination chain
        timeout_height (Height): Timeout height relative to the current block height.
            The timeout is disabled when set to 0.
        timeout_timestamp (int): Timeout timestamp (in nanoseconds) relative to the current block timestamp.
            The timeout is disabled when set to 0.
    """

    type = "cosmos-sdk/MsgTransfer"
    """"""
    type_url = "/ibc.applications.transfer.v1.MsgTransfer"
    """"""
    prototype = MsgTransfer_pb
    """"""

    source_port: str = attr.ib()
    source_channel: str = attr.ib()
    token: Coin = attr.ib(converter=Coin.parse)
    sender: AccAddress = attr.ib()
    receiver: str = attr.ib()  # stay str-typed because it may not be our address
    timeout_height: Height = attr.ib()
    timeout_timestamp: int = attr.ib(converter=int)

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "source_port": self.source_port,
                "source_channel": self.source_channel,
                "token": self.token.to_amino(),
                "sender": self.sender,
                "receiver": self.receiver,
                "timeout_height": self.timeout_height.to_amino(),
                "timeout_timestamp": self.timeout_timestamp,
            },
        }

    @classmethod
    def from_data(cls, data: dict) -> MsgTransfer:
        return cls(
            source_port=data["source_port"],
            source_channel=data["source_channel"],
            token=Coin.from_data(data["token"]),
            sender=data["sender"],
            receiver=data["receiver"],
            timeout_height=Height.from_data(data["timeout_height"]),
            timeout_timestamp=data["timeout_timestamp"],
        )

    def to_proto(self) -> MsgTransfer_pb:
        return MsgTransfer_pb(
            source_port=self.source_port,
            source_channel=self.source_channel,
            token=self.token.to_proto(),
            sender=self.sender,
            receiver=self.receiver,
            timeout_height=self.timeout_height.to_proto(),
            timeout_timestamp=self.timeout_timestamp,
        )

    @classmethod
    def from_proto(cls, proto: MsgTransfer_pb) -> MsgTransfer:
        return cls(
            source_port=proto.source_port,
            source_channel=proto.source_channel,
            token=Coin.from_proto(proto.token),
            sender=proto.sender,
            receiver=proto.receiver,
            timeout_height=Height.from_proto(proto.timeout_height),
            timeout_timestamp=proto.timeout_timestamp,
        )



import base64
import json
from typing import Any, List, Union

from terra_sdk.core import Numeric
from terra_sdk.core.wasm.data import AbsoluteTxPosition, HistoryEntry

from ._base import BaseAsyncAPI, sync_bind

__all__ = ["AsyncWasmAPI", "WasmAPI"]


class AsyncWasmAPI(BaseAsyncAPI):
    async def code_info(self, code_id: int) -> dict:
        """Fetches information about an uploaded code.

        Args:
            code_id (int): code ID

        Returns:
            dict: code information
        """
        res = await self._c._get(f"/cosmwasm/wasm/v1/code/{code_id}")
        code_info = res.get("code_info")
        return {
            "code_id": Numeric.parse(code_info["code_id"]),
            "data_hash": code_info["data_hash"],
            "creator": code_info["creator"],
            "instantiate_permission": code_info["instantiate_permission"],
        }

    async def contract_history(self, contract_address: str) -> List[HistoryEntry]:
        """Fetches contract history.

        Args:
            contract_address (str): contract address

        Returns:
            List[HistoryEntry]: contract histories
        """

        res = await self._c._get(
            f"/cosmwasm/wasm/v1/contract/{contract_address}/history"
        )

        return [HistoryEntry.from_data(entry) for entry in res["entries"]]

    async def contract_info(self, contract_address: str) -> dict:
        """Fetches information about an instantiated contract.

        Args:
            contract_address (str): contract address

        Returns:
            dict: contract information
        """
        res = await self._c._get(f"/cosmwasm/wasm/v1/contract/{contract_address}")
        contract_info = res.get("contract_info")
        contract_address = res.get("address")
        history_entries = self.contract_history(contract_address)
        return {
            "code_id": Numeric.parse(contract_info["code_id"]),
            "address": contract_address,
            "creator": contract_info["creator"],
            "admin": contract_info.get("admin", None),
            "label": contract_info.get("label", None),
            "init_msg": history_entries[0].msg,
            "created": AbsoluteTxPosition.from_data(contract_info.get("created"))
            if contract_info.get("created", None)
            else None,
            "ibc_port_id": contract_info.get("ibc_port_id", None),
        }

    async def contract_query(
        self, contract_address: str, query: Union[dict, str]
    ) -> Any:
        """Runs a QueryMsg on a contract.

        Args:
            contract_address (str): contract address
            query (dict): QueryMsg to run

        Returns:
            Any: results of query
        """
        query_msg = base64.b64encode(json.dumps(query).encode("utf-8")).decode("utf-8")
        res = await self._c._get(
            f"/cosmwasm/wasm/v1/contract/{contract_address}/smart/{query_msg}"
        )
        return res.get("data")

    async def pinned_codes(self) -> dict:
        """Fetches the Wasm module pinned codes.

        Returns:
            dict: Wasm module pinned codes
        """
        res = await self._c._get("/cosmwasm/wasm/v1/codes/pinned")
        params = res.get("code_ids")
        return {
            "code_ids": params,
        }


class WasmAPI(AsyncWasmAPI):
    @sync_bind(AsyncWasmAPI.code_info)
    def code_info(self, code_id: int) -> dict:
        pass

    code_info.__doc__ = AsyncWasmAPI.code_info.__doc__

    @sync_bind(AsyncWasmAPI.contract_info)
    def contract_info(self, contract_address: str) -> dict:
        pass

    contract_info.__doc__ = AsyncWasmAPI.contract_info.__doc__

    @sync_bind(AsyncWasmAPI.contract_query)
    def contract_query(self, contract_address: str, query: Union[dict, str]) -> Any:
        pass

    contract_query.__doc__ = AsyncWasmAPI.contract_query.__doc__

    @sync_bind(AsyncWasmAPI.pinned_codes)
    def pinned_codes(self) -> dict:
        pass

    pinned_codes.__doc__ = AsyncWasmAPI.pinned_codes.__doc__

    @sync_bind(AsyncWasmAPI.contract_history)
    def contract_history(self, contract_address: str) -> List[HistoryEntry]:
        pass

    contract_history.__doc__ = AsyncWasmAPI.contract_history.__doc__



from typing import Optional

from terra_sdk.core import AccAddress, Coins

from ..params import APIParams
from ._base import BaseAsyncAPI, sync_bind

__all__ = ["AsyncBankAPI", "BankAPI"]


class AsyncBankAPI(BaseAsyncAPI):
    async def balance(
        self, address: AccAddress, params: Optional[APIParams] = None
    ) -> (Coins, dict):
        """Fetches an account's current balance.

        Args:
            address (AccAddress): account address
            params (APIParams, optional): additional params for the API like pagination

        Returns:
            Coins: balance
            Pagination: pagination info
        """
        res = await self._c._get(f"/cosmos/bank/v1beta1/balances/{address}", params)
        return Coins.from_data(res["balances"]), res.get("pagination")

    async def total(self, params: Optional[APIParams] = None) -> (Coins, dict):
        """Fetches the current total supply of all tokens.

        Returns:
            Coins: total supply
            params (APIParams, optional): additional params for the API like pagination
        """
        res = await self._c._get("/cosmos/bank/v1beta1/supply", params)
        return Coins.from_data(res.get("supply")), res.get("pagination")

    async def spendable_balances(
        self, address: AccAddress, params: Optional[APIParams] = None
    ) -> (Coins, dict):
        """Queries the spenable balance of all coins for a single account

        Returns:
            Coins: spendable balance
            params (APIParams, optional): additional params for the API like pagination
        """
        res = await self._c._get(
            f"/cosmos/bank/v1beta1/spendable_balances/{address}", params
        )
        return Coins.from_data(res.get("balances")), res.get("pagination")


class BankAPI(AsyncBankAPI):
    @sync_bind(AsyncBankAPI.balance)
    def balance(
        self, address: AccAddress, params: Optional[APIParams] = None
    ) -> (Coins, dict):
        pass

    balance.__doc__ = AsyncBankAPI.balance.__doc__

    @sync_bind(AsyncBankAPI.total)
    def total(self) -> (Coins, dict):
        pass

    @sync_bind(AsyncBankAPI.spendable_balances)
    def spendable_balances(self) -> (Coins, dict):
        pass

    balance.__doc__ = AsyncBankAPI.balance.__doc__
    total.__doc__ = AsyncBankAPI.total.__doc__
    spendable_balances.__doc__ = AsyncBankAPI.spendable_balances.__doc__



from terra_sdk.core.public_key import PublicKey
from terra_sdk.core.tx import TxInfo, TxLog, parse_tx_logs

from .account import Account
from .base_account import BaseAccount
from .continuous_vesting_account import ContinuousVestingAccount
from .delayed_vesting_account import DelayedVestingAccount
from .periodic_vesting_account import PeriodicVestingAccount

__all__ = [
    "TxLog",
    "TxInfo",
    "PublicKey",
    "parse_tx_logs",
    "Account",
    "BaseAccount",
    "ContinuousVestingAccount",
    "DelayedVestingAccount",
    "PeriodicVestingAccount",
]



"""ibc channel module data objects."""
from __future__ import annotations

from typing import List

import attr
from terra_proto.ibc.core.channel.v1 import Channel as Channel_pb
from terra_proto.ibc.core.channel.v1 import Counterparty as Counterparty_pb
from terra_proto.ibc.core.channel.v1 import Order
from terra_proto.ibc.core.channel.v1 import Packet as Packet_pb
from terra_proto.ibc.core.channel.v1 import State

from terra_sdk.core.ibc.data.client import Height
from terra_sdk.util.json import JSONSerializable

__all__ = ["Counterparty", "Channel", "Order", "State", "Packet"]


@attr.s
class Counterparty(JSONSerializable):
    """
    Counterparty defines a channel end counterparty
    """

    port_id: str = attr.ib()
    channel_id: str = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> Counterparty:
        return cls(
            port_id=data["port_id"],
            channel_id=data["channel_id"],
        )

    def to_proto(self) -> Counterparty_pb:
        return Counterparty_pb(port_id=self.port_id, channel_id=self.channel_id)

    @classmethod
    def from_proto(cls, proto: Counterparty_pb) -> Counterparty:
        return cls(
            port_id=proto.port_id,
            channel_id=proto.channel_id,
        )


@attr.s
class Channel(JSONSerializable):
    """
    Channel defines pipeline for exactly-once packet delivery between specific
    modules on separate blockchains, which has at least one end capable of
    sending packets and one end capable of receiving packets.
    """

    state: State = attr.ib(converter=int)
    ordering: Order = attr.ib(converter=int)
    counterparty: Counterparty = attr.ib()
    connection_hops: List[str] = attr.ib(converter=list)
    version: str = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> Channel:
        return cls(
            state=data["state"],
            ordering=data["ordering"],
            counterparty=Counterparty.from_data(data["counterparty"]),
            connection_hops=data["connection_hops"],
            version=data["version"],
        )

    def to_proto(self) -> Channel_pb:
        return Channel_pb(
            state=self.state,
            ordering=self.ordering,
            counterparty=self.counterparty.to_proto(),
            connection_hops=self.connection_hops,
            version=self.version,
        )

    @classmethod
    def from_proto(cls, proto: Channel_pb) -> Channel:
        return cls(
            state=proto.state,
            ordering=proto.ordering,
            counterparty=Counterparty.from_proto(proto.counterparty),
            connection_hops=proto.connection_hops,
            version=proto.version,
        )


@attr.s
class Packet(JSONSerializable):
    """
    Packet defines a type that carries data across different chains through IBC
    """

    sequence: int = attr.ib(converter=int)
    source_port: str = attr.ib()
    source_channel: str = attr.ib()
    destination_port: str = attr.ib()
    destination_channel: str = attr.ib()
    data: bytes = attr.ib()
    timeout_height: Height = attr.ib()
    timeout_timestamp: int = attr.ib(converter=int)

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> Packet:
        return cls(
            sequence=data["sequence"],
            source_port=data["source_port"],
            source_channel=data["source_channel"],
            destination_port=data["destination_port"],
            destination_channel=data["destination_channel"],
            data=data["data"],
            timeout_height=Height.from_data(data["timeout_height"]),
            timeout_timestamp=data["timeout_timestamp"],
        )

    def to_proto(self) -> Packet_pb:
        return Packet_pb(
            sequence=self.sequence,
            source_port=self.source_port,
            source_channel=self.source_channel,
            destination_port=self.destination_port,
            destination_channel=self.destination_channel,
            data=self.data,
            timeout_height=self.timeout_height.to_proto(),
            timeout_timestamp=self.timeout_timestamp,
        )

    @classmethod
    def from_proto(cls, proto: Packet_pb) -> Packet:
        return cls(
            sequence=proto.sequence,
            source_port=proto.source_port,
            source_channel=proto.source_channel,
            destination_port=proto.destination_port,
            destination_channel=proto.destination_channel,
            data=proto.data,
            timeout_height=Height.from_proto(proto.timeout_height),
            timeout_timestamp=proto.timeout_timestamp,
        )
